from serra.transformers.transformer import Transformer
from pyspark.sql.functions import lit,max,min
from pyspark.sql import Window

class GetMaxOrMinTransformer(Transformer):
    """
    Test transformer to add a column to the dataframe with the maximum or minimum value from another column.

    :param columns_and_operations: A dictionary specifying the column to be transformed and the operation to be performed.
                                   The dictionary format should be {'input_column': 'aggregation_type'}.
                                   The output_column will contain the result of the aggregation operation.
    :param new_column_names: A dictionary specifying the names of the new columns to create for each aggregation operation.
                             The dictionary format should be {'aggregation_type': 'new_column_name'}.
    :param group_by_columns: A list of column names to group the DataFrame by.
    """

    def __init__(self, columns_and_operations, new_column_names, group_by_columns):
        self.columns_and_operations = columns_and_operations
        self.new_column_names = new_column_names
        self.group_by_columns = group_by_columns

    def transform(self, df):
        """
        Add a column with the maximum or minimum value to the DataFrame.

        :param df: The input DataFrame.
        :return: A new DataFrame with an additional column containing the maximum or minimum value.
        """
        window_spec = Window.partitionBy(self.group_by_columns)

        i = 0
        for input_col_name, agg_type in self.columns_and_operations.items():
            if agg_type == 'max':
                result_col = max(input_col_name).over(window_spec)
            elif agg_type == 'min':
                result_col = min(input_col_name).over(window_spec)
            
            df = df.withColumn(self.new_column_names[i], result_col)
            i+=1
        
        return df




from serra.python.base import PythonTransformer

class SampleTransformer(PythonTransformer):
    """
    Gets sample of size N from dataset.

    :param size: Sample size
    :param type: Type of sample (first/last N rows, 1 of every 100 rows, 1 in 100 chance to include each row)
    """

    def __init__(self, size, type):
        self.size = int(size)
        self.type = type

    def transform(self, df):
        type_map = {
            "First N rows": df.head,
            "Last N rows": df.tail,
            "Random sample": df.sample
            # "1 in 100 chance to include each row": randomsample
        }

        sample_function = type_map.get(self.type)

        df = sample_function(self.size)
        return df



from serra.python.base import PythonTransformer

class CoalesceTransformer(PythonTransformer):
    """
    Coalesces values across multiple columns into a single column.
    """

    def __init__(self, target_column, source_columns):
        """
        Initializes the CoalesceTransformer.

        :param target_column: The column to store the coalesced values.
        :param source_columns: A list of columns to coalesce into the target_column.
        """
        self.target_column = target_column
        self.source_columns = source_columns

    def transform(self, df):
        """
        Performs the coalescing operation on the dataframe.

        :param df: The dataframe to transform.
        :return: The transformed dataframe with coalesced values.
        """
        for column in self.source_columns:
            df[self.target_column] = df[self.target_column].combine_first(df[column])
        return df


# Running a specific job
import os
import sys
from datetime import datetime
from loguru import logger

from serra.config_parser import ConfigParser
from serra.utils import get_path_to_user_configs_folder, write_to_file
from serra.databricks import upload_wheel_to_bucket, restart_server
from serra.runners.graph_runner import run_job_with_graph
from serra.exceptions import SerraRunException
from serra.profile import SerraProfile

PACKAGE_PATH = os.path.dirname(os.path.dirname(__file__))

# Setup logger
logger.remove()  # Remove the default sink
logger.add(sink=sys.stdout, format="<green>{time}</green> - <level>{level}</level> - <cyan>{message}</cyan>", colorize=True)

def run_job(job_name, config_location):
    """
    You can either run the job with a file that is found locally, or is uploaded to an s3 bucket
    The s3 bucket is specified through the profiles.yml

    Returns a json representation of the run
    """

    assert config_location in ["local", 'aws']

    cf = None
    serra_profile = None

    if config_location == "local":
        user_configs_folder = get_path_to_user_configs_folder()
        config_path = f"{user_configs_folder}/{job_name}.yml"
        cf = ConfigParser.from_local_config(config_path)

        serra_profile = SerraProfile.from_yaml_path("./profiles.yml")
    elif config_location == "aws":
        config_path = f"{job_name}.yml"
        cf = ConfigParser.from_s3_config(config_path)
        raise NotImplementedError("Must grab profiles.yml from AWS as well")

    monitor = run_job_with_graph(cf, serra_profile)
    return monitor.to_dict()

def run_job_safely(job_name, config_location):
    """Wrapper of run job with exception catcher
    """
    result = None
    try:
        print('#####JOB RUN#####', datetime.now())
        result = run_job(job_name, config_location)
    except SerraRunException as s:
        logger.error(s)
    return result

def update_package():
    # create wheel
    # upload wheel to aws
    # tell databricks to delete all packages
    # restart server
    upload_wheel_to_bucket()
    restart_server()



import requests
from serra.python.base import PythonReader
import pandas as pd

class CmsReader(PythonReader):

    def __init__(self):
        pass
    
    def read(self):
        url = "https://data.cms.gov/provider-data/api/1/metastore/schemas/dataset/items"
        response = requests.get(url)
        if response.status_code != 200:
            return False
        cms_files = response.json()
        
        # Sort files by date
        sorted_files = sorted(cms_files, key=lambda x: x['issued'], reverse=True)
        # Remove files we are not interested in
        filtered_files = [file for file in sorted_files if file['title'] == 'Provider Information']
        sorted_files = filtered_files
        # get most recent file
        most_recent_file_info = sorted_files[0]
        file_id = most_recent_file_info['identifier']

        get_file_url = f"https://data.cms.gov/provider-data/api/1/datastore/query/{file_id}/0?offset=0&count=true&results=true&schema=true&keys=true&format=json&rowIds=false"
        file_specific_response = requests.get(get_file_url)
        if file_specific_response.status_code != 200:
            return False
        file_output = file_specific_response.json()
        list_of_rows = file_output['results']

        # Convert to pandas 
        df = pd.DataFrame(list_of_rows)
        return df




from pyspark.sql import functions as F
from serra.transformers.transformer import Transformer

class DateTruncTransformer(Transformer):
    """
    A transformer to truncate a timestamp column to a specified unit.

    :param timestamp_column: The name of the timestamp column to be truncated.
    :param trunc_unit: The unit for truncating the timestamp (e.g., 'day', 'month', 'year').
    :param output_column: The name of the new column to create with the truncated timestamps.
    """

    def __init__(self, timestamp_column, trunc_unit, output_column):
        self.timestamp_column = timestamp_column
        self.trunc_unit = trunc_unit
        self.output_column = output_column

    def transform(self, df):
        """
        Truncate the specified timestamp column to the specified unit.

        :param df: The input DataFrame to be transformed.
        :return: A new DataFrame with the truncated timestamp column.
        """
        dt = F.to_timestamp(self.timestamp_column, "yyyy-MM-dd HH:mm:ss")
        
        if self.trunc_unit == "day":
            truncated_time = F.date_trunc("day", dt)
        elif self.trunc_unit == "month":
            truncated_time = F.date_trunc("month", dt)
        elif self.trunc_unit == "year":
            truncated_time = F.date_trunc("year", dt)
        else:
            truncated_time = None
        
        return df.withColumn(self.output_column, truncated_time)



from pyspark.sql import functions as F

from serra.transformers.transformer import Transformer

class RenameColumnTransformer(Transformer):
    """
    A transformer to rename a column in a DataFrame.

    :param old_name: The name of the column to be renamed.
    :param new_name: The new name to be assigned to the column.
    """

    def __init__(self, old_name, new_name):
        self.old_name = old_name
        self.new_name = new_name

    def transform(self, df):
        """
        Rename a column in the DataFrame.

        :param df: The input DataFrame to be transformed.
        :return: A new DataFrame with the specified column renamed.
        """

        df = df.withColumn(
            self.new_name, F.col(self.old_name)
        )
        return df.drop(F.col(self.old_name))

    


from pyspark.sql import functions as F
from pyspark.sql.types import FloatType
from geopy.distance import geodesic
from serra.transformers.transformer import Transformer

class GeoDistanceTransformer(Transformer):
    """
    A transformer to calculate distances between pairs of geographical coordinates.

    :param start_column: The name of the column containing user coordinates.
    :param end_column: The name of the column containing facility coordinates.
    :param distance_km_col: The name of the column to store the calculated distance in kilometers.
    :param distance_mi_col: The name of the column to store the calculated distance in miles.
    """

    def __init__(self, start_column, end_column, output_column):
        self.start_column = start_column
        self.end_column = end_column
        self.output_column = output_column
        
    def transform(self, df):
        """
        Calculate distances between pairs of coordinates and add them to the DataFrame.

        :param df: The input DataFrame to be transformed.
        :return: A new DataFrame with calculated distances added as new columns.
        """
        # UDF to calculate distance in kilometers
        def calculate_distance_km(user_coords, facility_coords):
            user_lat, user_lon = map(float, user_coords.split(','))
            facility_lat, facility_lon = map(float, facility_coords.split(','))
            return geodesic((user_lat, user_lon), (facility_lat, facility_lon)).kilometers

        calculate_distance_km_udf = F.udf(calculate_distance_km, FloatType())

        # UDF to calculate distance in miles
        def calculate_distance_mi(user_coords, facility_coords):
            user_lat, user_lon = map(float, user_coords.split(','))
            facility_lat, facility_lon = map(float, facility_coords.split(','))
            return geodesic((user_lat, user_lon), (facility_lat, facility_lon)).miles

        calculate_distance_mi_udf = F.udf(calculate_distance_mi, FloatType())

        # Calculate distances and add them to the DataFrame
        df_with_distances = df.withColumn(
            self.output_column,
            calculate_distance_mi_udf(df[self.start_column], df[self.end_column])
        )

        return df_with_distances



class Block:
    def __init__(self, name):
        self.name = name
        self.downstream_blocks = []
        self.upstream_blocks = []

    def add_downstream_task(self, downstream_task_name):
        self.downstream_blocks.append(downstream_task_name)

    def add_upstream_task(self, upstream_task_name):
        self.upstream_blocks.append(upstream_task_name)

    def has_no_dependencies(self):
        return len(self.upstream_blocks) == 0

def get_result(block: Block, inputs):
    # Just for playing around
    result = f"[{block.name}"
    for input in inputs:
        result = result + "|" + input
    result = result + "]"
    return result

class BlockGraph:
    def __init__(self, blocks):
        self.blocks: list[Block] = blocks
        self.name_to_block = {}
        for block in blocks:
            self.name_to_block[block.name] = block

        self.outputs = {}

    def add_block(self, block_name, deps):
        block = Block(block_name)
        for dep in deps:
            block.add_upstream_task(dep)

        self.blocks.append(block)
        self.name_to_block[block_name] = block

    def find_entry_points(self):
        entry_names = []

        # Get blocks that have not been executed yet
        possible_blocks = []
        for block in self.blocks:
            if block.name not in self.outputs:
                possible_blocks.append(block)

        for block in possible_blocks:
            ok_to_run = True

            for dep in block.upstream_blocks:
                if dep not in self.outputs:
                    # We don't have the available inputs yet
                    ok_to_run = False
                    break
            
            if ok_to_run:
                entry_names.append(block.name)
        
        return entry_names
    

    def execute(self, block_name: str):

        block = self.name_to_block[block_name]

        inputs = []
        for input_block in block.upstream_blocks:
            # Grab the df from the upstream tasks
            inputs.append(self.outputs[input_block])

        result = get_result(block, inputs)
        self.outputs[block.name] = result

if __name__=="__main__":
    # Example of usage
    graph = BlockGraph([])
    graph.add_block("SALES", [])
    graph.add_block("RATINGS", [])
    graph.add_block("JOIN", ["SALES", "RATINGS"])
    graph.add_block("WRITE", ["JOIN"])

    order = []
    entry_points = graph.find_entry_points()
    while len(entry_points) != 0:
        # execute
        order.append(entry_points[0])
        graph.execute(entry_points[0])
        entry_points = graph.find_entry_points()

    print(order)


from pyspark.sql import DataFrame
from pyspark.sql import functions as F

from serra.transformers.transformer import Transformer
from serra.exceptions import SerraRunException

class SelectTransformer(Transformer):
    """
    A transformer to perform a SELECT operation on a DataFrame.

    :param columns: A list of column names to select from the DataFrame.
    :param distinct_column: Optional. The column for which distinct values should be retained.
    :param filter_expression: Optional. A filter expression to apply to the DataFrame.
    """

    def __init__(self, columns=None, distinct_column=None, filter_expression=None):
        self.columns = columns or []
        self.distinct_column = distinct_column
        self.filter_expression = filter_expression

    def transform(self, df: DataFrame) -> DataFrame:
        """
        Perform the SELECT operation on the DataFrame.

        :param df: The input DataFrame to be transformed.
        :return: A new DataFrame containing only the selected columns.
        :raises: SerraRunException if no columns are specified in the configuration
                 or if none of the specified columns exist in the DataFrame.
        """
        if not self.columns:
            raise SerraRunException("No columns specified in the configuration.")

        selected_columns = [F.col(col) for col in self.columns if col in df.columns]

        if not selected_columns:
            raise SerraRunException("None of the specified columns exist in the DataFrame.")

        df = df.select(*selected_columns)
        
        if self.distinct_column is not None:
            df = df.dropDuplicates(self.distinct_column)

        if self.filter_expression is not None:
            df = df.filter(F.expr(self.filter_expression))
        return df






from serra.python.base import PythonTransformer

# TODO: Add more operators >, <
class FilterRowsTransformer(PythonTransformer):
    """
    Drop rows that have nulls in specified column(s).

    :param column: The column to filter on
    :param operator: Comparison type (==, !=, >, <)
    :param value: The value to filter column on
    """

    def __init__(self, column, operator, value):
        self.column = column
        self.operator = operator
        self.value = value

    def transform(self, df):
        """
        First, get the column type and then convert the value to that specific type to ensure correctness while also keeping flexiblity on user input (ie. type = int, value = '130000', convert to 13000). This will help for later on where we can display operators as "equals" for strings and "=" for ints like Alteryx.
        Second, use the operator map to allow for flexible input on the operator side and then set the correct operator in code.
        Third, construct the filter query and use pandas query operator to execute it and return the filtered df. 
        """
        # Get column type, convert value
        column_type = str(df.dtypes[self.column])

        type_map = {
            "object": str,
            "int64": int,
            "float64": float
        }
        type_conversion = type_map.get(column_type)
    
        converted_value = type_conversion(self.value)
        if column_type == "object":
            converted_value = f"'{converted_value}'"

        # Get correct operator from dictionary, convert operator
        operator_map = {
            "Equals": "==",
            "Does not equal": "!="
        }
        operator = operator_map.get(self.operator)

        # Combine all things together into query, run and return filtered df
        query_str = f"`{self.column}` {operator} {converted_value}"

        return df.query(query_str)



from serra.readers import Reader

class S3Reader(Reader):
    """
    A class to read data from Amazon S3 and return a Spark DataFrame.

    :param config: A dictionary containing the configuration for the reader.
                   It should have the following keys:
                   - 'bucket_name': The name of the S3 bucket containing the file.
                   - 'file_path': The path to the file within the S3 bucket.
                   - 'file_type': The type of file to be read (e.g., 'csv', 'parquet').
    """

    def __init__(self, bucket_name, file_path, file_type, options):
        self.bucket_name = bucket_name
        self.file_path = file_path
        self.file_type = file_type
        self.options = options
    
    def read(self):
        """
        Read data from Amazon S3 and return a Spark DataFrame.

        :return: A Spark DataFrame containing the data read from the S3 bucket.
        :raises: Any exceptions that may occur during file reading or DataFrame creation.
        """
        spark = self.spark
        s3_url = f"s3a://{self.bucket_name}/{self.file_path}"
        df_read = spark.read

        # To specify options like header: true
        if self.options is not None:
            for option_key, option_value in self.options.items():
                df_read = df_read.option(option_key, option_value)
            
        df = df_read.format(self.file_type).load(s3_url)

        return df



from serra.python.base import PythonTransformer

class SortTransformer(PythonTransformer):
    """
    Create new columns of cumulative sums for given columns. Can group by and sort. 

    :param columns: Columns to calculate sums.
    :param group_by: Column(s) to group by.
    :param sort_by: Column(s) to sort by.
    """

    def __init__(self, sort_by, ascending):
        self.sort_by = sort_by
        self.ascending = ascending.lower() == "true"

    def transform(self, df):
        return df.sort_values(by = self.sort_by, ascending = self.ascending)



from serra.python.base import PythonTransformer

class JoinTransformer(PythonTransformer):
    """
    Join two datasets together (defaults to inner join). Can be used to union datasets. 

    :param right_df: Second dataframe in join
    :param left_df_key: First dataframe's join key
    :param right_df_key: Second dataframe's join key
    :param join_type: Type of join (inner, right, left, outer/union). Defaults to inner. 
    """

    def __init__(self, left_df, right_df, left_df_key, right_df_key, join_type = "inner"):
        self.left_df = left_df
        self.right_df = right_df
        self.left_df_key = left_df_key
        self.right_df_key = right_df_key
        self.join_type = join_type

    def transform(self):
        return self.left_df.merge(self.right_df, left_on=self.left_df_key, right_on=self.right_df_key, how=self.join_type)



from pyspark.sql import functions as F

from serra.transformers.transformer import Transformer

class JoinWithConditionTransformer(Transformer):
    """
    A transformer to join two DataFrames together based on a specified join condition.

    :param join_type: The type of join to perform. Currently only 'inner' join is supported.
    :param condition: The condition for the join.
    :param join_on: A dictionary where the keys are the table names (or DataFrame aliases)
                    and the values are the column names to join on for each table.
                    Example: {'table1': 'column1', 'table2': 'column2'}
    """

    def __init__(self, join_type, condition, join_on):
        self.join_type = join_type
        self.condition = condition
        self.join_on = join_on

    def transform(self, df1, df2):
        """
        Join two DataFrames together based on the specified join condition.

        :param df1: The first DataFrame to be joined.
        :param df2: The second DataFrame to be joined.
        :return: A new DataFrame resulting from the join operation.
        :raises: SerraRunException if the join condition columns do not match between the two DataFrames.
        """
        
        df1 = df1.alias('a').join(df2.alias('b'), F.expr(self.condition), self.join_type)

        return df1
    
    


from serra.python.base import PythonTransformer

from fuzzywuzzy import fuzz
import nltk
from nltk.util import ngrams
import pandas as pd

def remove_phrases(text, phrases_to_remove):
    for phrase in phrases_to_remove:
        text = text.replace(phrase, '')
    return text

class FuzzyMatchTransformer(PythonTransformer):
    """
    Drop rows that have nulls in specified column(s).

    :param columns: The columns to look for
    """

    def __init__(self, left_df, columns_to_match, partition):
        self.left_df = left_df # facilities
        self.columns_to_match = columns_to_match
        self.partition = partition
        
    def transform(self, right_df):

        if 'name' in self.columns_to_match:
            all_tokens = [token for sublist in self.left_df['tokenized_name'].tolist() + right_df['tokenized_name'].tolist() for token in sublist]

            trigrams = list(ngrams(all_tokens, 3))
            bigrams = list(ngrams(all_tokens, 2))

            trigram_freq = nltk.FreqDist(trigrams)
            bigram_freq = nltk.FreqDist(bigrams)

            phrases_to_remove = [' '.join(trigram) for trigram, freq in trigram_freq.most_common(3)]
            phrases_to_remove += [' '.join(bigram) for bigram, freq in bigram_freq.most_common(2)]

        matched_data = pd.DataFrame()

        # Iterate over each unique state
        for state in self.left_df[self.partition].unique():
            print(state)
            # Subset the data for the current state
            df_state_facilities = self.left_df[self.left_df[self.partition] == state].copy()
            df_state_provider_info = right_df[right_df[self.partition] == state].copy()

            # Step 1: Exact match on state and name
            for left_column, right_column in self.columns_to_match.items():
                exact_match_name = pd.merge(df_state_facilities, df_state_provider_info,
                                            left_on=[left_column],
                                            right_on=[right_column],
                                            how='inner',
                                            suffixes=('_fac', '_prov'))

                exact_match_name[left_column + '_match'] = 1

                matched_data = pd.concat([matched_data, exact_match_name], ignore_index=True)
                print(len(matched_data))

                # Remove matched rows
                df_state_facilities = df_state_facilities[~df_state_facilities[left_column].isin(exact_match_name[left_column])]
                df_state_provider_info = df_state_provider_info[~df_state_provider_info[right_column].isin(exact_match_name[right_column])]

            # Step 3: Fuzzy matching on address for remaining rows

            match_series_list = []

            for left_column, right_column in self.columns_to_match.items():
                for index, row in df_state_facilities.iterrows():
                    # Initialize a list to store scores for each provider row against the current facility row
                    address_scores = []

                    for index_prov, row_prov in df_state_provider_info.iterrows():
                        # Calculate fuzzy match score for the current column
                        score = fuzz.token_set_ratio(row[left_column], row_prov[right_column])
                        address_scores.append((index_prov, score))

                    # Proceed only if there are scores to evaluate
                    if address_scores:
                        # Find the max score and corresponding index
                        max_score_index, max_score = max(address_scores, key=lambda x: x[1])

                        # If the max score is above a threshold, consider it a match
                        
                        if max_score > 93:
                            match_dict = {
                                **row.to_dict(),
                                **df_state_provider_info.loc[max_score_index].to_dict(),
                                f"{left_column}_match": -1,  # Marking the match with a special flag
                                'fuzzy_score': max_score
                            }

                            match_series = pd.Series(match_dict, name=index)
                            match_series_list.append(match_series.to_frame().T)  # Convert Series to DataFrame and append

                            # Remove matched row from df_state_facilities and df_state_provider_info
                            df_state_facilities = df_state_facilities.drop(index)
                            df_state_provider_info = df_state_provider_info.drop(max_score_index)

            # After iterating through all columns and rows, check if there are any matches to append
            if match_series_list:  # Check if the list is not empty
                matched_data = pd.concat([matched_data] + match_series_list, ignore_index=True)
        return matched_data




# Entry point for serra command line tool
import sys
import click
import schedule, time

from serra.run import update_package, run_job_safely
from serra.databricks import create_job
from serra.utils import validate_workspace
from serra.config import PACKAGE_PATH
from serra.utils import copy_folder

@click.group()
def main():
    pass

@main.command(name="run")
@click.argument("job_name")
def cli_run_job_from_job_dir(job_name):
    """Run a specific job locally
    """
    validate_workspace()
    run_job_safely(job_name, "local")


@main.command(name="schedule")
@click.argument("job_name")
@click.argument("interval", type=int)
@click.argument("unit")
def cli_schedule_job(job_name, interval, unit):
    def job_scheduler():
        print('Running...')

        validate_workspace()
        run_job_safely(job_name, "local")

    if unit == 'seconds':
        schedule.every(interval).seconds.do(job_scheduler)
    elif unit == 'minutes':
        schedule.every(interval).minutes.do(job_scheduler)
    elif unit == 'daily':
        schedule.every(interval).days.do(job_scheduler)
    elif unit == 'weekly':
        schedule.every(interval).weeks.do(job_scheduler)

    job_scheduler()
    while True:
        schedule.run_pending()
        time.sleep(1)

@main.command(name="deploy")
@click.argument("job_name")
def cli_create_job(job_name):
    """Create a databricks job
    """
    validate_workspace()
    create_job(job_name)

@main.command(name="create")
@click.argument("local_path", type=click.Path(), default="./workspace")
def cli_create(local_path):
    """Create starter workspace folder"""
    source_folder = f"{PACKAGE_PATH}/data/workspace_example"
    copy_folder(source_folder, local_path)

@main.command(name="update_package")
def cli_update_package():
    """Uploads package to aws, and restarts databricks cluster
    """
    update_package()
    
# only for use by databricks cluster
# Did not use click because there were wierd traceback errors
def serra_databricks():
    assert len(sys.argv) == 2
    job_name = sys.argv[1]
    run_job_safely(job_name, 'aws')
    
if __name__ == '__main__':
  main()


from loguru import logger

from serra.config_parser import ConfigParser
from serra.utils import import_class
from serra.runners.ExecutionGraph import BlockGraph
from serra.python.runners.monitor import Monitor
from serra.python.base import PythonStep

def convert_name_to_full_python(class_name):
    if "Reader" in class_name:
        return f"serra.python.readers.{class_name}"
    elif "Writer" in class_name:
        return f"serra.python.writers.{class_name}"
    else:
        return f"serra.python.transformers.{class_name}"

# Returns instatiated reader,writer,transformer class with config already passed in
def get_configured_block_object(block_name, cf: ConfigParser) -> PythonStep:
    config = cf.get_config_for_block(block_name)
    class_name = cf.get_class_name_for_step(block_name)

    full_class_name = convert_name_to_full_python(class_name)
    block_class = import_class(full_class_name)
    configured_block_object = block_class.from_config(config)
    # configured_block_object = block_class(config)
    return configured_block_object

# TODO: Add execute method to all blocks to remove this design
def is_reader(block_name, cf: ConfigParser):
    class_name = cf.get_class_name_for_step(block_name)
    return "Reader" in class_name

def is_writer(block_name, cf: ConfigParser):
    class_name = cf.get_class_name_for_step(block_name)
    return "Writer" in class_name

def is_Transformer(block_name, cf: ConfigParser):
    class_name = cf.get_class_name_for_step(block_name)
    return "Transformer" in class_name

def get_order_of_execution(cf: ConfigParser):
    # Get all the blocks and initialize the objects
    block_names = cf.get_blocks()
    graph = BlockGraph([])
    for block_name in block_names:
        # obj = get_configured_block_object(block_name, cf)
        dependencies = cf.get_dependencies_for_block(block_name)
        graph.add_block(block_name, dependencies)

    # Now determine a possible order of execution
    order = []
    entry_points = graph.find_entry_points()
    while len(entry_points) != 0:
        order.append(entry_points[0])
        graph.execute(entry_points[0])
        entry_points = graph.find_entry_points()

    return order

def run_job_with_graph(cf: ConfigParser) -> Monitor:
    """
    This function allows us to run a job that contains blocks connected through the
    input_block parameter in each of the transformers and writers. The basic idea
    is to construct a graph of all the blocks connected to their dependencies. We then
    choose to "execute" the node on the graph that has no dependencies. Once it is executed,
    we prune that block from the graph. We continue until execution is complete.

    The df_map dict is used to store the intermediate dataframes that are created during the 
    course of this execution. For example, if a block called step_read produces a dataframe df1,
    that dataframe can be later accessed through df_map['step_read']. This allows blocks that will
    be executed later to access this dataframe and link together properly.

    TODO: Verify that all nodes are connected in graph
    TODO: Verify there are no circular dependencies
    """

    monitor = Monitor()

    ordered_block_names = get_order_of_execution(cf)
    logger.info(f"Decided order of execution: {ordered_block_names}")
    df_map = {}

    for block_name in ordered_block_names:
        logger.info(f"Executing step {block_name}")

        block_obj = get_configured_block_object(block_name, cf)

        if is_reader(block_name, cf):
            df = block_obj.read()

            # Set the dataframe reference in the map
            df_map[block_name] = df
        elif is_writer(block_name, cf):
            assert len(block_obj.dependencies) == 1
            
            input_block_name = block_obj.dependencies[0]
            input_df = df_map[input_block_name]
            block_obj.write(input_df)
        elif is_Transformer(block_name, cf):
            assert len(block_obj.dependencies) >= 1

            input_dfs = [df_map[dep] for dep in block_obj.dependencies]
            df = block_obj.transform(*input_dfs)

            # Set the dataframe reference in the map
            df_map[block_name] = df

        monitor.log_job_step(block_name, df)

        print(df)

    # response = monitor.to_dict()
    # for key, value in response.items():
    #     print(key)
    #     print(value)

    return monitor


from serra.base import Step
from abc import ABC, abstractmethod

class Reader(Step):
    """
    Reader base class for ingesting data
    Enforce read method
    """
    @abstractmethod
    def read(self, fmt, path, predicate=None):
        pass

    @classmethod
    def from_config(cls, config):
        return cls(**config)

    @property
    def dependencies(self):
        return []
    
    def read_with_spark(self, spark):
        self.spark = spark
        return self.read()



from pyspark.sql import functions as F

from serra.transformers.transformer import Transformer

class CastColumnsTransformer(Transformer):
    """
    A transformer to cast columns in a DataFrame to specified data types.

    :param columns_to_cast: A dictionary where the keys are the target column names
                            and the values are lists containing the source column name
                            and the target data type to which the source column will be cast.
                            Example: {'target_column': ['source_column', 'target_data_type']}
    """

    def __init__(self, columns_to_cast):
        self.columns_to_cast = columns_to_cast

    def transform(self, df):
        """
        Cast columns in the DataFrame to specified data types.

        :param df: The input DataFrame to be transformed.
        :return: A new DataFrame with the specified columns cast to the target data types.
        """
        for target, [source, target_data_type] in self.columns_to_cast.items():
            df = df.withColumn(target, F.col(source).cast(target_data_type))

        return df


from serra.python.base import PythonTransformer
import string
import re

class DataCleansingTransformer(PythonTransformer):
    """
    Clean whitespaces, cases, punctuation.

    """

    def __init__(self, columns, case, punctuation=None, whitespace=None):
        self.columns = columns
        self.case = case
        self.punc = punctuation
        self.white_space = whitespace
        
    def transform(self, df):
        """
        Get mapping dictionaries. Keys are whatever use case you want, Values are dictionaries with keys as column name and values as the functions that do that specific use case transform.
        Loop through key-value pairs for the specified dictionaries, assign df[key] = value.

        """

        # Using dictionaries so we can extend later on ie) adding more punctuation cases/white space cases
        # Mappings — Key (type of case), Value (dictionary that has pairs where key is column name, value is function that outputs transformed column)
        case_map = {
            "Upper Case": {col: df[col].str.upper() for col in self.columns},
            "Lower Case": {col: df[col].str.lower() for col in self.columns}
        }
        
        # Iterate through the specified key to get dictionary and iterate through key value pairs to apply the transforms
        for col, cased_col in case_map.get(self.case, {}).items():
            df[col] = cased_col

        white_space_map = {
            "Leading and Trailing Whitespace": {col: df[col].str.strip() for col in self.columns},
            "All Whitespace": {col: df[col].str.replace(" ", "", regex=False) for col in self.columns},
        }

        for col, white_space_col in white_space_map.get(self.white_space, {}).items():
            df[col] = white_space_col

        punc_map = {
            "Punctuation": {col: df[col].str.replace(f"[{re.escape(string.punctuation)}]", "", regex=True) for col in self.columns},
        }

        for col, punc_col in punc_map.get(self.punc, {}).items():
            df[col] = punc_col

        return df





from pyspark.sql import functions as F

from serra.transformers.transformer import Transformer

class CoalesceTransformer(Transformer):
    """
    A transformer to create a new column by coalescing multiple columns.

    :param input_columns: A list of column names to coalesce.
    :param output_column: The name of the new column to create with the coalesced values.
    """

    def __init__(self, input_columns, output_column):
        self.input_columns = input_columns
        self.output_column = output_column

    def transform(self, df):
        """
        Create a new column by coalescing multiple columns.

        :param df: The input DataFrame to be transformed.
        :return: A new DataFrame with the coalesced column.
        """

        return df.withColumn(self.output_column, F.coalesce(*self.input_columns))



from serra.readers import Reader
from serra.exceptions import SerraRunException

class DatabricksReader(Reader):
    """
    A reader to read data from a Databricks Delta Lake table into a Spark DataFrame.

    :param config: A dictionary containing the configuration for the reader.
                   It should have the following keys:
                   - 'database': The name of the database containing the table.
                   - 'table': The name of the table to be read.
    """

    def __init__(self, database, table):
        self.database = database
        self.table = table

    def read(self):
        """
        Read data from a Databricks Delta Lake table and return a Spark DataFrame.

        :return: A Spark DataFrame containing the data read from the specified table.
        :raises: SerraRunException if an error occurs during the data reading process.
        """
        try:
            df = self.spark.read.table(f'{self.database}.{self.table}')
        except Exception as e:
            raise SerraRunException(e)
        return df



from .reader import Reader
from .local_reader import LocalReader
from .databricks_reader import DatabricksReader
from .s3_reader import S3Reader
from .snowflaker_reader import SnowflakeReader
from .big_query_reader import BigQueryReader
from .mongo_db_reader import MongoDBReader




import pandas as pd
from abc import ABC, abstractmethod

class PythonStep(ABC):
    pass

class PythonReader(PythonStep):
    """
    Reader base class for ingesting data
    Enforce read method
    """
    @abstractmethod
    def read(self) -> pd.DataFrame:
        pass

    @classmethod
    def from_config(cls, config):
        c = dict(config)
        obj = cls(**c)
        return obj

    @property
    def dependencies(self):
        return []

class PythonTransformer(PythonStep):
    """
    Transformer base class for data transformations
    Enforce transform method
    """
    @abstractmethod
    def transform(self, df: pd.DataFrame):
        pass

    @classmethod
    def from_config(cls, config):
        c = dict(config)
        input_block = c.pop("input_block")

        obj = cls(**c)
        obj.input_block = input_block
        return obj

    @property
    def dependencies(self):
        return [self.input_block]
    
class PythonWriter(PythonStep):
    """
    Writer base class for loading data
    Enforce write method
    """
    @abstractmethod
    def write(self, df: pd.DataFrame):
        pass

    @classmethod
    def from_config(cls, config):
        c = dict(config)
        input_block = c.pop("input_block")

        obj = cls(**c)
        obj.input_block = input_block
        return obj

    @property
    def dependencies(self):
        return [self.input_block]


from pyspark.sql import functions as F

from serra.transformers.transformer import Transformer
from serra.exceptions import SerraRunException

class PivotTransformer(Transformer):
    """
    A transformer to pivot a DataFrame based on specified row and column levels, and perform aggregation.

    :param row_level_column: The column used for row levels during pivoting.
    :param column_level_column: The column used for column levels during pivoting.
    :param value_column: The column to be summarized (values) during pivoting.
    :param aggregate_type: The type of aggregation to perform after pivoting.
                           Should be one of 'avg' (average) or 'sum' (sum).
    """

    def __init__(self, row_level_column, column_level_column, value_column, aggregate_type):
        self.row_level_column = row_level_column
        self.column_level_column = column_level_column
        self.value_column = value_column
        self.aggregate_type = aggregate_type

    def transform(self, df):
        """
        Pivot the DataFrame based on the specified row and column levels, and perform aggregation.

        :param df: The input DataFrame to be transformed.
        :return: A new DataFrame resulting from the pivot operation with the specified aggregation.
        :raises: SerraRunException if the specified aggregation type is invalid.
        """
        
        df = df.withColumn(self.value_column, F.col(self.value_column).cast("double"))
        df = df.groupBy(self.row_level_column).pivot(self.column_level_column)

        # Perform aggregation
        if self.aggregate_type == "avg":
            df = df.avg(self.value_column)
        elif self.aggregate_type == "sum":
            df = df.sum(self.value_column)
        else:
            raise SerraRunException("Invalid Pivot Aggregation type")

        return df


    


from pyspark.sql import functions as F

from serra.transformers.transformer import Transformer

## helpers to execute custom transform code
import re
import importlib

def get_user_transform(code_string, desired_function_name):
    # Extract and perform imports
    imports = re.findall(r"from ([\w\.]+) import ([\w ,]+)", code_string)
    for imp in imports:
        module_name, components = imp
        module = importlib.import_module(module_name)
        for component in components.split(","):
            component = component.strip()
            if " as " in component:
                actual_name, alias = component.split(" as ")
                globals()[alias] = getattr(module, actual_name)
            else:
                globals()[component] = getattr(module, component)

    # Execute the provided code string
    local_namespace = {}
    exec(code_string, globals(), local_namespace)
    
    # Rename the transform function
    globals()[desired_function_name] = local_namespace["transform"]

class CustomTransformer(Transformer):

    def __init__(self, code_block):
        self.code_block = code_block

    def transform(self, df):
        custom_transform_func_name = "custom_transform_" + str(id(self))
        get_user_transform(self.code_block, custom_transform_func_name)
        output_df = globals()[custom_transform_func_name](df)
        return output_df


from serra.python.base import PythonTransformer
import pandas as pd

class CountRowsTransformer(PythonTransformer):
    """
    Get the row count for the dataset.

    """

    def __init__(self):
        pass
        
    def transform(self, df):
        row_count = len(df.index)
        df = pd.DataFrame({'Count': [row_count]})
        return df



from pyspark.sql import DataFrame
from pyspark.sql import functions as F
from pyspark.sql import Window
from serra.transformers.transformer import Transformer

class WindowTransformer(Transformer):
    """
    A transformer to set a window for all other steps after it.

    :param partition_by: A list of column names for partitioning the window.
    :param order_by: A list of column names for ordering within the window.
    :param window_name: The name of the window column to be added.
    """

    def __init__(self, partition_by=None, order_by=None, window_name=None):
        self.partition_by = partition_by or []
        self.order_by = order_by or []
        self.window_name = window_name

    def transform(self, df: DataFrame) -> DataFrame:
        window_spec = Window().partitionBy(*self.partition_by).orderBy(*self.partition_by)
        df_with_window = df.withColumn(f"window_{self.partition_by[0]}", F.row_number().over(window_spec))
        return df_with_window


import requests
import os

# Constants
BASE_URL = "********"
TOKEN = '******'
HEADERS = {
    "Authorization": TOKEN
}
SCRIPT = "py_script.py"
DBFS_PATH = "dbfs:/scripts/t.py"

def upload_script_to_dbfs(script_path, dbfs_path):
    """Uploads a script to DBFS."""
    with open(script_path, "rb") as f:
        script_content = f.read()

    response = requests.post(
        f"{BASE_URL}/api/2.0/dbfs/put",
        headers=HEADERS,
        files={
            'path': (None, dbfs_path),
            'content': (os.path.basename(script_path), script_content),
            'overwrite': (None, 'true')
        }
    )

    if response.status_code == 200:
        print(f"Script uploaded successfully to {dbfs_path}")
    else:
        print(f"Error uploading script: {response.content}")
        return False

    return True

def create_and_run_job(dbfs_path):
    """Creates and runs a Databricks job."""
    # Create a new job
    response = requests.post(
        f"{BASE_URL}/api/2.0/jobs/create",
        headers=HEADERS,
        json={
            "name": "MyJob3",
            "existing_cluster_id": "0922-180514-m6fcc4kt",
            "spark_python_task": {
                "python_file": dbfs_path
            }
        }
    )
    assert response.status_code == 200
    job_id = response.json()["job_id"]

    # Run the job
    response = requests.post(
        f"{BASE_URL}/api/2.0/jobs/run-now",
        headers=HEADERS,
        json={"job_id": job_id}
    )
    assert response.status_code == 200

if __name__ == "__main__":
    if upload_script_to_dbfs(SCRIPT, DBFS_PATH):
        create_and_run_job(DBFS_PATH)



# Helper functions
import numpy as np
def get_json_rep_pandas(df, rows=20):
    df_head = df.head(rows).replace({np.nan: None, float('nan'): None})

    # Get column names
    columns = list(df.columns)

    # Get schema information
    schema_info = [{"name": col, "type": str(df[col].dtype)} for col in df.columns]

    # Convert each row to a dictionary and collect them as a list
    json_data = df_head.to_dict('records')

    # Create a dictionary for the JSON output
    output_dict = {
        "columns": columns,
        "schema": schema_info,
        "data": json_data
    }
    return output_dict

# Output helpers
class Monitor:
    # PythonMonitor
    def __init__(self):
        self.step_to_df = {}
        self.step_to_json_rep = {}

    def to_dict(self):
        # Provide a json representation of the run
        ret = {
            "json_rep": self.step_to_json_rep
        }
        return ret
    
    
    def log_job_step(self,step_name, df):
        self.step_to_json_rep[step_name] = get_json_rep_pandas(df)


from pyspark.sql import functions as F

from serra.transformers.transformer import Transformer
from serra.exceptions import SerraRunException

class AddColumnTransformer(Transformer):
    """
    A transformer to add a new column to the DataFrame with a specified value.

    :param new_column_name: The name of the new column to be added.
    :param value: The value to be set for the new column.
    :param new_column_type: The data type of the new column. Must be a valid PySpark data type string.
    """

    def __init__(self, new_column_name, value, new_column_type):
        self.new_column_name = new_column_name
        self.value = value
        self.new_column_type = new_column_type
        
    def transform(self, df):
        """
        Add a new column to the DataFrame with the specified name, value, and data type.

        :param df: The input DataFrame to be transformed.
        :return: A new DataFrame with the new column added.
        :raises: SerraRunException if the column with the specified name already exists in the DataFrame.
        """
        if self.new_column_name in df.columns:
            raise SerraRunException(f"Column '{self.new_column_name}' already exists in the DataFrame. Choose a different name.")
        
        return df.withColumn(
            self.new_column_name, F.lit(self.value).cast(self.new_column_type)  
        )



from serra.python.base import PythonReader
import pandas as pd

class LocalReader(PythonReader):
    def __init__(self, path: str):
        self.path = path
    
    def read(self):
        df = pd.read_csv(self.path)
        return df


from pyspark.sql.functions import when, count, isnull, col

from serra.exceptions import SerraRunException

def nulls_test(df):
    has_nulls = df.select([count(when(isnull(c), c)).alias(c) for c in df.columns])
    if has_nulls:
        raise SerraRunException("Null values were found in the DataFrame")
    
def duplicates_test(df):
    if df.count() > df.dropDuplicates(df.columns).count():
        raise SerraRunException('Data has duplicates')
    
def check_dates(df):
    # Define the timestamp columns
    # Define the timestamp column
    timestamp_column = "time"

    # Apply the date comparison to the specified column
    df = df.withColumn(timestamp_column, col(timestamp_column) > "2020-01-01T00:00:00Z")
    # Use 'all' to check if all columns satisfy the condition
    # Count the rows where the condition is satisfied
    count_true = df.filter(col(timestamp_column)).count()

    # Compare the count with the total number of rows
    if count_true != df.count():
        raise SerraRunException('Data has dates outside of valid range')




from serra.transformers import Transformer
from pyspark.sql import Window
from pyspark.sql import functions as F

class DeDuplicateTransformer(Transformer):
    def __init__(self, columns_to_compare, time_frame):
        self.col_to_compare = columns_to_compare
        self.columns_to_compare = columns_to_compare
        self.time_frame = int(time_frame)

    def transform(self, df):
        df = df.repartition(*self.columns_to_compare)
        df = df.withColumn("event_time", F.to_timestamp("event_time"))
        window_spec = Window.partitionBy(self.columns_to_compare).orderBy("event_time")
        df = df.withColumn("rn", F.row_number().over(window_spec))
        df = df.withColumn("row_diff",
                    F.unix_timestamp(F.col("event_time")) - F.lag(F.unix_timestamp(F.col("event_time")), 1).over(window_spec))
        time_diff = 60 * self.time_frame
        df = df.filter((F.col("row_diff") > time_diff) | F.col("row_diff").isNull())
        df = df.drop("rn", "row_diff")

        return df



import os
import yaml

from serra.aws import retrieve_file_from_config_bucket
from serra.exceptions import SerraRunException

def convert_name_to_full(class_name):
    if "Reader" in class_name:
        return f"serra.readers.{class_name}"
    elif "Writer" in class_name:
        return f"serra.writers.{class_name}"
    else:
        return f"serra.transformers.{class_name}"

class ConfigParser:

    def __init__(self, config):
        self.config = config
    
    @staticmethod
    def from_string(config_string):
        config = yaml.safe_load(config_string)
        return ConfigParser(config)

    @staticmethod
    def from_local_config(config_path: str):
        if (not os.path.isfile(config_path)):
            raise SerraRunException(f"Unable to find config file: {config_path}")

        with open(config_path, 'r') as stream:
            config = yaml.safe_load(stream)
        return ConfigParser(config)
    
    @staticmethod
    def from_s3_config(config_name):
        # TODO: Make more generalizable
        # Currenltly 
        config_bytes = retrieve_file_from_config_bucket(config_name)
        config = yaml.safe_load(config_bytes)
        return ConfigParser(config)
    
    def get_step(self, block_name):
        return self.config.get(block_name)
    
    def get_metadata_tags(self):
        return ['show_all']
    
    def get_class_name_for_step(self, block_name):
        step = self.get_step(block_name)
        keys = [key for key in step.keys()]
        if "tests" in keys:
            keys.remove("tests")
        return keys[0]
    
    def get_config_for_step(self, block_name):
        step = self.get_step(block_name)
        return step.get("config")
    
    def get_blocks(self):
        blocks = [name for name in self.config.keys() if name != 'show_all']
        return blocks
    
    def get_config_for_block(self, block_name):
        class_name = self.get_class_name_for_step(block_name)
        return self.config.get(block_name).get(class_name)
    
    def get_tests_for_block(self, block_name):
        return self.config.get(block_name).get("tests")
    
    def show_all(self):
        return self.config.get('show_all')
    
    def get_dependencies_for_block(self, block_name):
        block_config = self.get_config_for_block(block_name)
        result = block_config.get("input_block", [])
        if (type(result) == str):
            return [result]
        return result



from serra.writers.writer import Writer

class LocalWriter(Writer):
    """
    A writer to write data from a Spark DataFrame to a local file.

    :param file_path: The path of the local file to write to.
    """

    def __init__(self, file_path):
        self.file_path = file_path

    def write(self, df):
        """
        Write data from a Spark DataFrame to a local file.

        :param df: The Spark DataFrame to be written to the local file.
        """
        # Convert PySpark DataFrame to Pandas DataFrame
        pandas_df = df.toPandas()

        # Write the Pandas DataFrame to a local file
        pandas_df.to_csv(self.file_path, index=False)


from serra.readers import Reader
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.session import SparkSession

class MongoDBReader(Reader):
    def __init__(self, username, password, database, collection, cluster_ip_and_options):
        self.username = username
        self.password = password
        self.database = database
        self.collection = collection
        self.cluster_ip_and_options = cluster_ip_and_options

    def read(self):
        spark = self.spark

        #TODO: connection.uri can either start iwth mongodb+srv or mongodb
        """
        Cluster must have this package installed:
        org.mongodb.spark:mongo-spark-connector_2.12:10.2.1

        Must also run Spark 3.1 - 3.2.4 according to https://www.mongodb.com/docs/spark-connector/v10.2/
        """

        df = spark.read\
        .format("mongodb")\
        .option("connection.uri", f'mongodb+srv://{self.username}:{self.password}@{self.cluster_ip_and_options}')\
        .option("database", self.database)\
        .option("collection", self.collection)\
        .load()
        return df


from pyspark.sql import functions as F

from serra.exceptions import SerraRunException
from serra.transformers.transformer import Transformer

class CrossJoinTransformer(Transformer):
    """
    A transformer to join two DataFrames together based on a specified join condition.

    :param config: A dictionary containing the configuration for the transformer.
                   It should have the following keys:
                   - 'join_type': The type of join to perform. Currently only 'inner' join is supported.
                   - 'join_on': A dictionary where the keys are the table names (or DataFrame aliases)
                                and the values are the column names to join on for each table.
                   Example: {'table1': 'column1', 'table2': 'column2'}
    """

    def __init__(self):
        pass

    @property
    def dependencies(self):
        return self.config.get('input_block')

    def transform(self, df1, df2):
        """
        Join two DataFrames together based on the specified join condition.

        :param df1: The first DataFrame to be joined.
        :param df2: The second DataFrame to be joined.
        :return: A new DataFrame resulting from the join operation.
        :raises: SerraRunException if the join condition columns do not match between the two DataFrames.
        """
        
        return df1.alias('a').crossJoin(df2.alias('b'))
    
    


# A helper class to keep track of the run and provide results

def getShowString(df, n=20, truncate=True, vertical=False):
    if isinstance(truncate, bool) and truncate:
        return(df._jdf.showString(n, 20, vertical))
    else:
        return(df._jdf.showString(n, int(truncate), vertical))

def get_json_rep(df, rows=20):
    df_head = df.limit(20)

    # Get column names
    columns = df.columns

    # Get schema information
    schema_info = [field.jsonValue() for field in df.schema.fields]

    # Convert each row to a dictionary and collect them as a list
    json_data = []
    for row in df_head.collect():
        row_dict = {col: str(value) for col, value in zip(columns, row)}
        json_data.append(row_dict)

    # Create a dictionary for the JSON output
    output_dict = {
        "columns": columns,
        "schema": schema_info,
        "data": json_data
    }
    return output_dict


class Monitor:
    def __init__(self):
        self.step_to_df = {}
        self.step_to_json_rep = {}
        pass

    def to_dict(self):
        # Provide a json representation of the run
        ret = {
            "string_rep": self.step_to_df,
            "json_rep": self.step_to_json_rep
        }
        return ret
    
    
    def log_job_step(self,step_name, df):
        self.step_to_df[step_name] = getShowString(df)
        self.step_to_json_rep[step_name] = get_json_rep(df)



from pyspark.sql.functions import col, when

from serra.transformers.transformer import Transformer

class MultipleCaseWhenTransformer(Transformer):
    """
    A transformer that adds a new column to the DataFrame based on conditional rules (similar to SQL's CASE WHEN).

    :param columns_and_conditions: A list of tuples representing the conditions and their corresponding results.
                                   Each tuple should be in the format (condition_value, result_value).
                                   The condition_value can be a specific value or a pattern for LIKE comparisons.
                                   The result_value will be assigned to the output_col if the condition is met.
    :param input_column: The name of the column containing the values to be evaluated.
    :param output_col: The name of the new column to be added with the results of the conditions.
    :param type: The type of comparison to be used. It can be either '==' for equality comparison or 'like' for pattern matching using the LIKE operator.
    """

    def __init__(self, columns_and_conditions, input_column, output_col, type):
        self.columns_and_conditions = columns_and_conditions
        self.input_column = input_column
        self.output_col = output_col
        self.type = type

    def transform(self, df):
        """
        Add a new column with the results of the conditions to the DataFrame.

        :param df: The input DataFrame.
        :return: A new DataFrame with an additional column containing the results of the conditions.
        """
        # Mappings to indicate which comparison function to use
        type_dict = {
            '<=': lambda col, val: col <= val,
            '>=': lambda col, val: col >= val,
            '==': lambda col, val: col == val,
            'like': lambda col, val: col.like(val)
        }

        # Get the particular condition function based on the comparison type
        comparison_func = type_dict.get(self.type)
        output_cols = list(self.columns_and_conditions.keys())
        conditions = list(self.columns_and_conditions.values())

        if comparison_func is None:
            raise ValueError(f"Unsupported comparison type: {self.type}")
        
        for i in range(len(output_cols)):
            if len(conditions[i]) == 2:
                case_expr = when(comparison_func(df[self.input_column[0]], conditions[i][0]), self.parse_result_value(conditions[i][1]))
            else: 
                case_expr = when(comparison_func(df[self.input_column[0]], conditions[i][0]) | comparison_func(df[self.input_column[0]], conditions[i][2]), self.parse_result_value(conditions[i][1]))
            case_expr = case_expr.otherwise(None)
            df = df.withColumn(output_cols[i], case_expr)

        return df
    
    def parse_result_value(self, result_value):
        if 'col:' in result_value:
            return col(result_value[4:])
        else:
            return(result_value)



# Module to help connect to databricks
from loguru import logger
from databricks.sdk import WorkspaceClient
from databricks.sdk.service.jobs import Task, PythonWheelTask
from databricks.sdk.service.compute import Library, State

from serra.aws import upload_file_to_config_bucket
from serra.utils import get_path_to_user_configs_folder
from serra.utils import get_local_serra_profile
from serra.config import (
    WHEEL_FILE_NAME_IN_BUCKET,
    LOCAL_PATH_TO_WHEEL
)

def get_remote_path_to_wheel():
    serra_profile = get_local_serra_profile()
    bucket = serra_profile.aws_config_bucket
    s3_wheel_path = f"s3://{bucket}/{WHEEL_FILE_NAME_IN_BUCKET}"
    return s3_wheel_path

def create_databricks_workspace_client():
    serra_profile = get_local_serra_profile()
    w = WorkspaceClient(
        host  = serra_profile.databricks_host,
        token = serra_profile.databricks_token
    )
    return w

def upload_wheel_to_bucket():
    # TODO: Add code to recreate wheel
    logger.info(f"Uploading wheel from {LOCAL_PATH_TO_WHEEL} to AWS")
    upload_file_to_config_bucket(LOCAL_PATH_TO_WHEEL)

def create_job(config_name):
    """Use like this: submit_job("StripeExample")
    """
    serra_profile = get_local_serra_profile()
    s3_wheel_path = get_remote_path_to_wheel()

    w = create_databricks_workspace_client()

    # Upload the config file to aws
    logger.info("Uploading config to AWS")
    config_path = f"{get_path_to_user_configs_folder()}/{config_name}.yml"
    upload_file_to_config_bucket(config_path)
    
    # Create the job on databricks
    logger.info("Creating databricks job")
    job_name = config_name

    j = w.jobs.create(
        name=job_name,
        tasks = [
            Task(
                description = "Basic new job",
                existing_cluster_id = serra_profile.databricks_cluster_id,
                python_wheel_task=PythonWheelTask(entry_point="serra_databricks",
                                                package_name="serra",
                                                parameters=[config_name]),
                libraries=[Library(whl=s3_wheel_path)],
                task_key="my-task"
            )
    ]
    )
    logger.info(f"View job at {w.config.host}/#job/{j.job_id}")
    return j

def restart_server():
    w = create_databricks_workspace_client()
    serra_profile = get_local_serra_profile()
    s3_wheel_path = get_remote_path_to_wheel()
    cluster_id = serra_profile.databricks_cluster_id

    logger.info("Restarting cluster")
    w.clusters.ensure_cluster_is_running(cluster_id)

    logger.info("Uninstalling packages")
    w.libraries.uninstall(cluster_id=cluster_id,
                          libraries=[Library(whl=s3_wheel_path)])
    
    current_cluster_state = w.clusters.get(cluster_id).state

    if current_cluster_state != State.TERMINATED:
        w.clusters.restart_and_wait(cluster_id)
    
    w.clusters.ensure_cluster_is_running(cluster_id)

    logger.info("Finished restarting")


from serra.transformers import Transformer
from pyspark.sql import DataFrame

class DropNullTransformer(Transformer):
    """
    A transformer to drop rows with null values in specified columns.

    :param columns: A list of column names to check for null values.
    """

    def __init__(self, columns):
        self.columns = columns

    def transform(self, df: DataFrame) -> DataFrame:
        """
        Drop rows with null values in specified columns.

        :param df: The input DataFrame to be transformed.
        :return: A new DataFrame with rows containing null values dropped.
        """
        for column in self.columns:
            df = df.filter(df[column].isNotNull())
        return df


from serra.python.base import PythonTransformer

class DropNullRowsTransformer(PythonTransformer):
    """
    Drop rows that have nulls in specified column(s).

    :param columns: The columns to look for
    """

    def __init__(self, columns):
        self.columns = columns
        
    def transform(self, df):
        """
        Add a new column to the DataFrame with the specified name, value, and data type.

        :param df: The input DataFrame to be transformed.
        :return: A new DataFrame with the new column added.
        :raises: SerraRunException if the column with the specified name already exists in the DataFrame.
        """
        
        return df.dropna(subset=self.columns)



from serra.python.base import PythonTransformer

class UniqueTransformer(PythonTransformer):
    """
    Select specified columns

    :param columns: The columns to look for duplicate values in
    """

    def __init__(self, columns):
        self.columns = columns

    def transform(self, df):
        """
        If columns = [restaurant,id], then we want to drop rows that have both the same restaurant and id value

        """
        return df.drop_duplicates(self.columns)



import pandas as pd

from serra.python.base import PythonTransformer

class TransposeTransformer(PythonTransformer):
    """
    Gets sample of size N from dataset.

    :param size: Sample size
    :param type: Type of sample (first/last N rows, 1 of every 100 rows, 1 in 100 chance to include each row)
    """

    def __init__(self, columns):
        self.columns = columns

    def transform(self, df):
        df = pd.melt(df[self.columns],var_name='Name',value_name='Value')
        return df



import yaml

class SerraProfile:

    def __init__(self, config):
        self.config = config

    @staticmethod
    def from_string(config_string):
        config = yaml.safe_load(config_string)
        return SerraProfile(config)

    @staticmethod
    def from_yaml_path(config_path):
        with open(config_path, 'r') as stream:
            config = yaml.safe_load(stream)
        return SerraProfile(config)

    @property
    def aws_access_key_id(self):
        return self.config.get("AWS_ACCESS_KEY_ID")
    
    @property
    def aws_secret_access_key(self):
        return self.config.get("AWS_SECRET_ACCESS_KEY")
    
    @property
    def aws_config_bucket(self):
        return self.config.get("AWS_CONFIG_BUCKET")
    
    @property
    def databricks_host(self):
        return self.config.get("DATABRICKS_HOST")
    
    @property
    def databricks_token(self):
        return self.config.get("DATABRICKS_TOKEN")
    
    @property
    def databricks_cluster_id(self):
        return self.config.get("DATABRICKS_CLUSTER_ID")
    
    @property
    def snowflake_account(self):
        return self.config.get("SNOWFLAKE")




import os
import shutil

from os.path import exists
from loguru import logger
from serra.profile import SerraProfile

def get_local_serra_profile():
    serra_profile = SerraProfile.from_yaml_path("./profiles.yml")
    return serra_profile

def write_to_file(filename, content):
    try:
        with open(filename, 'w') as file:
            file.write(content)
        logger.info(f"Successfully wrote to {filename}.")
    except IOError:
        logger.error(f"Error writing to {filename}.")

def import_class(cl):
    d = cl.rfind(".")
    classname = cl[d + 1 : len(cl)]
    m = __import__(cl[0:d], globals(), locals(), [classname])
    return getattr(m, classname)

def get_path_to_user_configs_folder():
    return "./jobs"

def read_sql_file(file_path):
    with open(file_path, 'r') as file:
        sql_script = file.read()
    return sql_script

def validate_workspace():
    if not exists("./jobs"):
        logger.error("Please create a jobs directory before proceeding")
        exit()
    if not exists("./profiles.yml"):
        logger.error("Please create a profiles.yml file in the current directory.")
        exit()
    return

def copy_folder(source_folder, destination_folder):
    try:
        # Check if the source folder exists
        if not os.path.exists(source_folder):
            logger.info(f"Source folder '{source_folder}' does not exist.")
            return

        # Check if the destination folder exists
        if not os.path.exists(destination_folder):
            os.makedirs(destination_folder)

        # Copy the contents of the source folder to the destination folder
        for item in os.listdir(source_folder):
            source_item = os.path.join(source_folder, item)
            destination_item = os.path.join(destination_folder, item)
            if os.path.isdir(source_item):
                shutil.copytree(source_item, destination_item)
            else:
                shutil.copy2(source_item, destination_item)

        logger.info(f"Folder copied successfully from '{source_folder}' to '{destination_folder}'.")
    except Exception as e:
        logger.error(f"An error occurred while copying the folder: {e}")

def dict_to_string(input_dict):
    # Initialize an empty list to store key-value pairs as strings
    key_value_strings = []

    for key, value in input_dict.items():
        # Check if the value is a string and wrap it in double quotes
        if key == 'input_block':
            continue

        if isinstance(value, str):
            value_str = f'"{value}"'
        else:
            value_str = value  # Keep other types as they are

        # Remove single quotes from the key and format the key-value pair
        key_str = key.strip("'")  # Remove single quotes if present
        key_value_strings.append(f'{key_str} = {value_str}')

    # Combine the key-value strings into a single string
    result = ", ".join(key_value_strings)

    return result


from serra.base import Step
from abc import abstractmethod

class Transformer(Step):
    """
    Transformer base class for data transformations
    Enforce transform method
    """
    @abstractmethod
    def transform(self, df):
        pass

    @classmethod
    def from_config(cls, config):
        c = dict(config)
        input_block = c.pop("input_block")

        obj = cls(**c)
        obj.input_block = input_block
        return obj

    @property
    def dependencies(self):
        return [self.input_block]


from pyspark.sql.functions import col,when 

from serra.transformers.transformer import Transformer

class CaseWhenTransformer(Transformer):
    """
    Test transformer to add a column to dataframe
    :param output_column: The name of the new output column to be added.
    :param input_column: The name of the input column.
    :param conditions: A list of conditions to evaluate. Each condition should be a tuple of (condition_column, operator, value).
    :param comparison_type: The type of comparison to use (e.g., 'equals', 'greater_than').
    :param is_column_condition: Boolean indicating whether the condition is based on a column value.
    :param otherwise_value: The value to use when none of the conditions are met.
    """

    def __init__(self, output_column, input_column, conditions, comparison_type, is_column_condition, otherwise_value):
        self.output_column = output_column
        self.input_column = input_column
        self.conditions = conditions
        self.comparison_type = comparison_type
        self.is_column_condition = is_column_condition
        self.otherwise_value = otherwise_value

    def transform(self, df):
        """
        Add column with col_value to dataframe
        :return; Dataframe w/ new column containing col_value
        """
        print("Conditions:", self.conditions[0][0])
        # Mappings to indicate which comparison function to use
        type_dict = {
            '<=': lambda col, val: col <= val,
            '>=': lambda col, val: col >= val,
            '==': lambda col, val: col == val,
            'like': lambda col, val: col.like(val)
        }

        # Get the particular condition function based on the comparison type
        comparison_func = type_dict.get(self.comparison_type)

        if comparison_func is None:
            raise ValueError(f"Unsupported comparison type: {self.comparison_type}")
        
        # Create the 'when' expression based on the provided conditions and type
        if self.is_column_condition is None:
            case_expr = when(comparison_func(df[self.input_column], self.conditions[0][0]), self.parse_result_value(self.conditions[0][1]))
            for cond_val, result_val in self.conditions[1:]:
                case_expr = case_expr.when(comparison_func(df[self.input_column], cond_val), self.parse_result_value(result_val))

            # Apply the 'otherwise' function to specify the default value (None for the last condition)
            case_expr = case_expr.otherwise(None)

            return df.withColumn(self.output_column, case_expr)
        else:
            case_expr = when(comparison_func(df[self.input_column], df[self.conditions[0][0]]), self.parse_result_value(self.conditions[0][1]))
            for cond_val, result_val in self.conditions[1:]:
                case_expr = case_expr.when(comparison_func(df[self.input_column], cond_val), self.parse_result_value(result_val))

            # Apply the 'otherwise' function to specify the default value (None for the last condition)
            case_expr = case_expr.otherwise(self.otherwise_value)

            return df.withColumn(self.output_column, case_expr)
    
    def parse_result_value(self, result_value):
        if isinstance(result_value,str) and 'col:' in result_value:
            return col(result_value[4:])
        else:
            return(result_value)



from pyspark.sql import functions as F

from serra.transformers.transformer import Transformer

class GetCountTransformer(Transformer):
    """
    A transformer to get the count of occurrences of a column in the DataFrame.

    :param group_by_columns: A list of column names to group the DataFrame by.
    :param count_column: The name of the column for which the count is to be calculated.
    """

    def __init__(self, group_by_columns, count_column):
        self.group_by_columns = group_by_columns
        self.count_column = count_column

    def transform(self, df):
        """
        Calculate the count of occurrences of the specified column in the DataFrame.

        :param df: The input DataFrame to be transformed.
        :return: A new DataFrame with the count of occurrences of the specified column
                 for each group defined by the 'group_by' column(s).
        """

        return df.groupBy(*self.group_by_columns).agg(F.count(self.count_column))
    


from google.cloud import bigquery

from serra.exceptions import SerraRunException
from serra.writers import Writer

class BigQueryWriter(Writer):
    """
    A writer to write data to BigQuery from a Spark DataFrame.

    :param config: A dictionary containing the configuration for the reader.
                   It should have the following keys:
                   - 'project_id'
                   - 'dataset_id'
                   - 'table_id'
    """

    def __init__(self, project_id, dataset_id, table_id, mode):
        self.project_id = project_id
        self.dataset_id = dataset_id
        self.table_id = table_id
        self.mode = mode
    
    def write(self, df):
        """
        Read data from Snowflake and return a Spark DataFrame.

        :return: A Spark DataFrame containing the data read from the specified Snowflake table.
        """
        df.write \
            .format("bigquery") \
            .option('project', self.project_id)\
            .option("writeMethod", "direct") \
            .mode(self.mode)\
            .save(f"{self.dataset_id}.{self.table_id}")



import nltk
from nltk.tokenize import word_tokenize
from nltk.corpus import stopwords
from serra.python.base import PythonTransformer

class TokenizeTransformer(PythonTransformer):
    """
    Tokenize your strings.
    """

    def __init__(self, columns):
        self.columns = columns
        
    def transform(self, df):
        """
        Get mapping dictionaries. Keys are whatever use case you want, Values are dictionaries with keys as column name and values as the functions that do that specific use case transform.
        Loop through key-value pairs for the specified dictionaries, assign df[key] = value.

        """

        stop_words = set(stopwords.words('english'))
        df[['tokenized_' + col for col in self.columns]] = df[self.columns].applymap(lambda x: [word for word in word_tokenize(x) if word not in stop_words])

        return df





from serra.python.base import PythonTransformer

class SelectColumnsTransformer(PythonTransformer):
    """
    Select specified columns

    :param columns: The columns to select
    """

    def __init__(self, columns):
        self.columns = columns

    @classmethod
    def from_config(cls, config):
        columns = config.get('columns')

        obj = cls(columns)
        return obj

    def transform(self, df):
        return df[self.columns]



from .local_writer import LocalWriter
from .writer import Writer
from .databricks_writer import DatabricksWriter
from .s3_writer import S3Writer
from .snowflake_writer import SnowflakeWriter
from .big_query_writer import BigQueryWriter



from serra.writers import Writer

class S3Writer(Writer):
    """
    A writer to write data from a Spark DataFrame to Amazon S3 as a CSV file.

    :param config: A dictionary containing the configuration for the writer.
                   It should have the following keys:
                   - 'bucket_name': The name of the S3 bucket to write to.
                   - 'file_path': The path to the CSV file in the S3 bucket.
                   - 'file_type': The file type, which should be 'csv' for this writer.
                   - 'input_block': The name of the input block used as a dependency for this writer.
    """
    
    def __init__(self, bucket_name, file_path, file_type, options, mode):
        self.bucket_name = bucket_name
        self.file_path = file_path
        self.file_type = file_type
        self.options = options
        self.mode = mode
    
    def write(self, df):
        """
        Write data from a Spark DataFrame to Amazon S3 as a CSV file.

        :param df: The Spark DataFrame to be written.
        """
        df_write = df.write

        # To specify options like header: true
        if self.options is not None:
            for option_key, option_value in self.options.items():
                df_write = df_write.option(option_key, option_value)
        
        s3_url = f"s3a://{self.bucket_name}/{self.file_path}"
        df_write.mode(self.mode).format(self.file_type).save(s3_url)



from pyspark.sql import functions as F

from serra.transformers.transformer import Transformer

class GetDistinctTransformer(Transformer):
    """
    A transformer to drop duplicate rows from the DataFrame based on specified column(s).

    :param columns_to_check: A list of column names to identify rows for duplicates removal.
    """

    def __init__(self, columns_to_check):
        self.columns_to_check = columns_to_check

    def transform(self, df):
        """
        Drop duplicate rows from the DataFrame based on specified columns.

        :param df: The input DataFrame to be transformed.
        :return: A new DataFrame with duplicate rows removed based on the specified columns.
        """

        return df.dropDuplicates(self.columns_to_check)



from pyspark.sql import SparkSession
from abc import ABC, abstractmethod
from pyspark.sql.functions import col, when
import yaml
import json
import os
from pyspark.sql import functions as F
from pyspark.sql.types import FloatType
from pyspark.sql.functions import lit,max,min
from pyspark.sql import Window
from pyspark.ml.feature import Imputer
from pyspark.sql import DataFrame

class SerraProfile:

    def __init__(self, config):
        self.config = config

    @staticmethod
    def from_string(config_string):
        config = yaml.safe_load(config_string)
        return SerraProfile(config)

    @staticmethod
    def from_yaml_path(config_path):
        with open(config_path, 'r') as stream:
            config = yaml.safe_load(stream)
        return SerraProfile(config)

    @property
    def aws_access_key_id(self):
        return self.config.get("AWS_ACCESS_KEY_ID")
    
    @property
    def aws_secret_access_key(self):
        return self.config.get("AWS_SECRET_ACCESS_KEY")
    
    @property
    def aws_config_bucket(self):
        return self.config.get("AWS_CONFIG_BUCKET")
    
    @property
    def databricks_host(self):
        return self.config.get("DATABRICKS_HOST")
    
    @property
    def databricks_token(self):
        return self.config.get("DATABRICKS_TOKEN")
    
    @property
    def databricks_cluster_id(self):
        return self.config.get("DATABRICKS_CLUSTER_ID")
    
    @property
    def snowflake_account(self):
        return self.config.get("SNOWFLAKE")

class Step(ABC):
    def add_spark_session(self, spark):
        self.spark = spark

class Reader(Step):
    """
    Reader base class for ingesting data
    Enforce read method
    """
    @abstractmethod
    def read(self, fmt, path, predicate=None):
        pass

    @property
    def dependencies(self):
        return []

class Transformer(Step):
    """
    Transformer base class for data transformations
    Enforce transform method
    """
    @abstractmethod
    def transform(self, df):
        pass

    @property
    def dependencies(self):
        return [self.input_block]

class Writer(Step):
    """
    Writer base class for loading data
    Enforce write method
    """
    @abstractmethod
    def write(self, df):
        pass

    @property
    def dependencies(self):
        return [self.input_block]

class SerraRunException(Exception):
    "Raised when error is detected when running a Serra job"
    pass

class LocalReader(Reader):
    """
    A reader to read data from a local file into a Spark DataFrame.

    :param config: A dictionary containing the configuration for the reader.
                   It should have the following key:
                   - 'file_path': The path to the local file to be read.
    """


    def __init__(self, file_path):
        self.file_path = file_path

    @classmethod
    def from_config(cls, config):
        file_path = config.get("file_path")
        return cls(file_path)
        
    def read(self):
        """
        Read data from a local file and return a Spark DataFrame.

        :return: A Spark DataFrame containing the data read from the local file.
        """
        #TODO: check all files supports
        df = self.spark.read.format("csv").option("header",True).load(self.file_path)
        return df

    def read_with_spark(self, spark):
        self.spark = spark
        return self.read()
    
class DatabricksReader(Reader):
    """
    A reader to read data from a Databricks Delta Lake table into a Spark DataFrame.

    :param config: A dictionary containing the configuration for the reader.
                   It should have the following keys:
                   - 'database': The name of the database containing the table.
                   - 'table': The name of the table to be read.
    """

    def __init__(self, database, table):
        self.database = database
        self.table = table
        
    @classmethod
    def from_config(cls, config):
        database = config.get('database')
        table = config.get('table')

        obj = cls(database, table)
        return obj

    def read(self):
        """
        Read data from a Databricks Delta Lake table and return a Spark DataFrame.

        :return: A Spark DataFrame containing the data read from the specified table.
        :raises: SerraRunException if an error occurs during the data reading process.
        """
        try:
            df = self.spark.read.table(f'{self.database}.{self.table}')
        except Exception as e:
            raise SerraRunException(e)
        return df

    def read_with_spark(self, spark):
        self.spark = spark
        return self.read()


class JoinTransformer(Transformer):
    """
    A transformer to join two DataFrames together based on a specified join condition.

    :param config: A dictionary containing the configuration for the transformer.
                   It should have the following keys:
                   - 'join_type': The type of join to perform. Currently only 'inner' join is supported.
                   - 'join_on': A dictionary where the keys are the table names (or DataFrame aliases)
                                and the values are the column names to join on for each table.
                   Example: {'table1': 'column1', 'table2': 'column2'}
    """


    def __init__(self, join_type, join_on):
        self.join_type = join_type
        self.join_on = join_on

    @classmethod
    def from_config(cls, config):
        join_type = config.get("join_type")
        join_on = config.get("join_on")
        obj = cls(join_type, join_on)
        obj.input_block = config.get("input_block")
        return obj
    
    @property
    def dependencies(self):
        return self.input_block

    def transform(self, df1, df2):
        """
        Join two DataFrames together based on the specified join condition.

        :param df1: The first DataFrame to be joined.
        :param df2: The second DataFrame to be joined.
        :return: A new DataFrame resulting from the join operation.
        :raises: SerraRunException if the join condition columns do not match between the two DataFrames.
        """
        # assert self.join_type in "inner"

        join_keys = []

        
        for table in self.join_on:
            join_keys.append(self.join_on.get(table))

        assert len(join_keys) == 2

        matching_col = join_keys

        df1 = df1.join(df2, df1[matching_col[0]] == df2[matching_col[1]], self.join_type).drop(df2[matching_col[1]])
        if df1.isEmpty():
            raise SerraRunException(f"""Joiner - Join Key Error: {matching_col[0]}, {matching_col[1]} columns do not match.""")
        return df1 
    
class AddColumnTransformer(Transformer):
    """
    A transformer to add a new column to the DataFrame with a specified value.

    :param new_column_name: The name of the new column to be added.
    :param value: The value to be set for the new column.
    :param new_column_type: The data type of the new column. Must be a valid PySpark data type string.
    """

    def __init__(self, new_column_name, value, new_column_type):
        self.new_column_name = new_column_name
        self.value = value
        self.new_column_type = new_column_type

    @classmethod
    def from_config(cls, config):
        new_column_name = config.get("new_column_name")
        value = config.get("value")
        new_column_type = config.get("new_column_type")

        obj = cls(new_column_name, value, new_column_type)
        obj.input_block = config.get('input_block')
        return obj
        
    def transform(self, df):
        """
        Add a new column to the DataFrame with the specified name, value, and data type.

        :param df: The input DataFrame to be transformed.
        :return: A new DataFrame with the new column added.
        :raises: SerraRunException if the column with the specified name already exists in the DataFrame.
        """
        if self.new_column_name in df.columns:
            raise SerraRunException(f"Column '{self.new_column_name}' already exists in the DataFrame. Choose a different name.")
        
        return df.withColumn(
            self.new_column_name, F.lit(self.value).cast(self.new_column_type)  
        )

class AggregateTransformer(Transformer):
    """
    A transformer to aggregate data based on specified columns and aggregation type.

    :param group_by_columns: A list of column names to group the DataFrame by.
    :param aggregation_type: The type of aggregation to be performed. Supported values: 'sum', 'avg', 'count'.
    """

    def __init__(self, group_by_columns, aggregation_type):
        self.group_by_columns = group_by_columns
        self.aggregation_type = aggregation_type

    @classmethod
    def from_config(cls, config):
        group_by_columns = config.get('group_by_columns')
        aggregation_type = config.get('aggregation_type')

        obj = cls(group_by_columns, aggregation_type)
        obj.input_block = config.get('input_block')
        return obj

    def transform(self, df):
        """
        Transform the DataFrame by aggregating data based on the specified configuration.

        :param df: The input DataFrame to be transformed.
        :return: A new DataFrame with the aggregated data.
        """
        df = df.groupBy(*self.group_by_columns)

        if self.aggregation_type == 'sum':
            df = df.sum()
        if self.type == 'avg':
            df = df.mean()
        if self.type == 'count':
            df = df.count()

        return df

class CaseWhenTransformer(Transformer):
    """
    Test transformer to add a column to dataframe
    :param output_column: The name of the new output column to be added.
    :param input_column: The name of the input column.
    :param conditions: A list of conditions to evaluate. Each condition should be a tuple of (condition_column, operator, value).
    :param comparison_type: The type of comparison to use (e.g., 'equals', 'greater_than').
    :param is_column_condition: Boolean indicating whether the condition is based on a column value.
    :param otherwise_value: The value to use when none of the conditions are met.
    """

    def __init__(self, output_column, input_column, conditions, comparison_type, is_column_condition, otherwise_value):
        self.output_column = output_column
        self.input_column = input_column
        self.conditions = conditions
        self.comparison_type = comparison_type
        self.is_column_condition = is_column_condition
        self.otherwise_value = otherwise_value

    @classmethod
    def from_config(cls, config):
        output_column = config.get("output_column")
        input_column = config.get("input_column")
        conditions = config.get('conditions')
        comparison_type = config.get('comparison_type')
        is_column_condition = config.get('is_column_condition')
        otherwise_value = config.get('otherwise_value')

        obj = cls(output_column, input_column, conditions, comparison_type, is_column_condition, otherwise_value)
        obj.input_block = config.get('input_block')
        return obj

    def transform(self, df):
        """
        Add column with col_value to dataframe
        :return; Dataframe w/ new column containing col_value
        """
        print("Conditions:", self.conditions[0][0])
        # Mappings to indicate which comparison function to use
        type_dict = {
            '<=': lambda col, val: col <= val,
            '>=': lambda col, val: col >= val,
            '==': lambda col, val: col == val,
            'like': lambda col, val: col.like(val)
        }

        # Get the particular condition function based on the comparison type
        comparison_func = type_dict.get(self.comparison_type)

        if comparison_func is None:
            raise ValueError(f"Unsupported comparison type: {self.comparison_type}")
        print('###########', self.conditions[0][0])
        # Create the 'when' expression based on the provided conditions and type
        if self.is_column_condition is None:
            case_expr = when(comparison_func(df[self.input_column], self.conditions[0][0]), self.parse_result_value(self.conditions[0][1]))
            for cond_val, result_val in self.conditions[1:]:
                case_expr = case_expr.when(comparison_func(df[self.input_column], cond_val), self.parse_result_value(result_val))

            # Apply the 'otherwise' function to specify the default value (None for the last condition)
            case_expr = case_expr.otherwise(None)

            return df.withColumn(self.output_column, case_expr)
        else:
            case_expr = when(comparison_func(df[self.input_column], df[self.conditions[0][0]]), self.parse_result_value(self.conditions[0][1]))
            for cond_val, result_val in self.conditions[1:]:
                case_expr = case_expr.when(comparison_func(df[self.input_column], cond_val), self.parse_result_value(result_val))

            # Apply the 'otherwise' function to specify the default value (None for the last condition)
            case_expr = case_expr.otherwise(self.otherwise_value)

            return df.withColumn(self.output_column, case_expr)
    
    def parse_result_value(self, result_value):
        if isinstance(result_value,str) and 'col:' in result_value:
            return col(result_value[4:])
        else:
            return(result_value)

class CastColumnsTransformer(Transformer):
    """
    A transformer to cast columns in a DataFrame to specified data types.

    :param columns_to_cast: A dictionary where the keys are the target column names
                            and the values are lists containing the source column name
                            and the target data type to which the source column will be cast.
                            Example: {'target_column': ['source_column', 'target_data_type']}
    """

    def __init__(self, columns_to_cast):
        self.columns_to_cast = columns_to_cast

    @classmethod
    def from_config(cls, config):
        columns_to_cast = config.get("columns_to_cast")
        obj = cls(columns_to_cast)
        obj.input_block = config.get('input_block')
        return obj

    def transform(self, df):
        """
        Cast columns in the DataFrame to specified data types.

        :param df: The input DataFrame to be transformed.
        :return: A new DataFrame with the specified columns cast to the target data types.
        """
        for target, [source, target_data_type] in self.columns_to_cast.items():
            df = df.withColumn(target, F.col(source).cast(target_data_type))

        return df

class CoalesceTransformer(Transformer):
    """
    A transformer to create a new column by coalescing multiple columns.

    :param input_columns: A list of column names to coalesce.
    :param output_column: The name of the new column to create with the coalesced values.
    """

    def __init__(self, input_columns, output_column):
        self.input_columns = input_columns
        self.output_column = output_column

    @classmethod
    def from_config(cls, config):
        input_columns = config.get('input_columns')
        output_column = config.get('output_column')

        obj = cls(input_columns, output_column)
        obj.input_block = config.get('input_block')
        return obj

    def transform(self, df):
        """
        Create a new column by coalescing multiple columns.

        :param df: The input DataFrame to be transformed.
        :return: A new DataFrame with the coalesced column.
        """

        return df.withColumn(self.output_column, F.coalesce(*self.input_columns))

class CrossJoinTransformer(Transformer):
    """
    A transformer to join two DataFrames together based on a specified join condition.

    :param config: A dictionary containing the configuration for the transformer.
                   It should have the following keys:
                   - 'join_type': The type of join to perform. Currently only 'inner' join is supported.
                   - 'join_on': A dictionary where the keys are the table names (or DataFrame aliases)
                                and the values are the column names to join on for each table.
                   Example: {'table1': 'column1', 'table2': 'column2'}
    """

    def __init__(self):
        self = self

    @classmethod
    def from_config(cls, config):
        obj = cls()
        obj.input_block = config.get('input_block')
        return obj

    # @property
    # def dependencies(self):
    #     """
    #     Get the list of table names that this transformer depends on.

    #     :return: A list of table names (keys from the 'join_on' dictionary).
    #     """
    #     return [key for key in self.config.get("join_on").keys()]
    @property
    def dependencies(self):
        return self.config.get('input_block')

    def transform(self, df1, df2):
        """
        Join two DataFrames together based on the specified join condition.

        :param df1: The first DataFrame to be joined.
        :param df2: The second DataFrame to be joined.
        :return: A new DataFrame resulting from the join operation.
        :raises: SerraRunException if the join condition columns do not match between the two DataFrames.
        """
        
        return df1.alias('a').crossJoin(df2.alias('b'))

class DateTruncTransformer(Transformer):
    """
    A transformer to truncate a timestamp column to a specified unit.

    :param timestamp_column: The name of the timestamp column to be truncated.
    :param trunc_unit: The unit for truncating the timestamp (e.g., 'day', 'month', 'year').
    :param output_column: The name of the new column to create with the truncated timestamps.
    """

    def __init__(self, timestamp_column, trunc_unit, output_column):
        self.timestamp_column = timestamp_column
        self.trunc_unit = trunc_unit
        self.output_column = output_column

    @classmethod
    def from_config(cls, config):
        timestamp_column = config.get("timestamp_column")
        trunc_unit = config.get("trunc_unit")
        output_column = config.get('output_column')

        obj = cls(timestamp_column, trunc_unit, output_column)
        obj.input_block = config.get('input_block')
        return obj

    def transform(self, df):
        """
        Truncate the specified timestamp column to the specified unit.

        :param df: The input DataFrame to be transformed.
        :return: A new DataFrame with the truncated timestamp column.
        """
        dt = F.to_timestamp(self.timestamp_column, "yyyy-MM-dd HH:mm:ss")
        
        if self.trunc_unit == "day":
            truncated_time = F.date_trunc("day", dt)
        elif self.trunc_unit == "month":
            truncated_time = F.date_trunc("month", dt)
        elif self.trunc_unit == "year":
            truncated_time = F.date_trunc("year", dt)
        else:
            truncated_time = None
        
        return df.withColumn(self.output_column, truncated_time)

class DropColumnsTransformer(Transformer):
    """
    A transformer to drop specified columns from the DataFrame.

    :param config: A dictionary containing the configuration for the transformer.
                   It should have the following key:
                   - 'columns_to_drop': A list of column names to be dropped from the DataFrame.
    """

    def __init__(self, columns_to_drop):
        self.columns_to_drop = columns_to_drop

    @classmethod
    def from_config(cls, config):
        columns_to_drop = config.get("columns_to_drop")
        obj = cls(columns_to_drop)
        obj.input_block = config.get('input_block')
        return obj

    def transform(self, df):
        """
        Drop specified columns from the DataFrame.

        :param df: The input DataFrame to be transformed.
        :return: A new DataFrame with the specified columns dropped.
        """
        return df.select([c for c in df.columns if c not in self.columns_to_drop])

class FilterTransformer(Transformer):
    """
    A transformer to filter the DataFrame based on a specified condition.

    :param filter_values: A list of values to filter the DataFrame by.
    :param filter_column: The name of the column to apply the filter on.
    :param is_expression: Whether the filter is an expression (boolean condition).
    """

    def __init__(self, filter_values, filter_column, is_expression=False):
        self.filter_values = filter_values
        self.filter_column = filter_column
        self.is_expression = is_expression

    @classmethod
    def from_config(cls, config):
        filter_values = config.get("filter_values")
        filter_column = config.get("filter_column")
        is_expression = config.get("is_expression", False)
        obj = cls(filter_values, filter_column, is_expression)
        obj.input_block = config.get('input_block')
        return obj
        

    def transform(self, df):
        """
        Filter the DataFrame based on the specified condition.

        :param df: The input DataFrame to be transformed.
        :return: A new DataFrame with rows filtered based on the specified condition.
        """
        if self.is_expression:
            return df.filter(F.expr(self.filter_values))

        return df.filter(df[self.filter_column].isin(self.filter_values))

class GetCountTransformer(Transformer):
    """
    A transformer to get the count of occurrences of a column in the DataFrame.

    :param group_by_columns: A list of column names to group the DataFrame by.
    :param count_column: The name of the column for which the count is to be calculated.
    """

    def __init__(self, group_by_columns, count_column):
        self.group_by_columns = group_by_columns
        self.count_column = count_column

    @classmethod
    def from_config(cls, config):
        group_by_columns = config.get("group_by_columns")
        count_column = config.get("count_column")

        obj = cls(group_by_columns, count_column)
        obj.input_block = config.get('input_block')
        return obj

    def transform(self, df):
        """
        Calculate the count of occurrences of the specified column in the DataFrame.

        :param df: The input DataFrame to be transformed.
        :return: A new DataFrame with the count of occurrences of the specified column
                 for each group defined by the 'group_by' column(s).
        """

        return df.groupBy(*self.group_by_columns).agg(F.count(self.count_column))

class GetDistinctTransformer(Transformer):
    """
    A transformer to drop duplicate rows from the DataFrame based on specified column(s).

    :param columns_to_check: A list of column names to identify rows for duplicates removal.
    """

    def __init__(self, columns_to_check):
        self.columns_to_check = columns_to_check

    @classmethod
    def from_config(cls, config):
        columns_to_check = config.get('columns_to_check')

        obj = cls(columns_to_check)
        obj.input_block = config.get('input_block')
        return obj

    def transform(self, df):
        """
        Drop duplicate rows from the DataFrame based on specified columns.

        :param df: The input DataFrame to be transformed.
        :return: A new DataFrame with duplicate rows removed based on the specified columns.
        """

        return df.dropDuplicates(self.columns_to_check)

class GetMaxOrMinTransformer(Transformer):
    """
    Test transformer to add a column to the dataframe with the maximum or minimum value from another column.

    :param columns_and_operations: A dictionary specifying the column to be transformed and the operation to be performed.
                                   The dictionary format should be {'input_column': 'aggregation_type'}.
                                   The output_column will contain the result of the aggregation operation.
    :param new_column_names: A dictionary specifying the names of the new columns to create for each aggregation operation.
                             The dictionary format should be {'aggregation_type': 'new_column_name'}.
    :param group_by_columns: A list of column names to group the DataFrame by.
    """

    def __init__(self, columns_and_operations, new_column_names, group_by_columns):
        self.columns_and_operations = columns_and_operations
        self.new_column_names = new_column_names
        self.group_by_columns = group_by_columns

    @classmethod
    def from_config(cls, config):
        columns_and_operations = config.get('columns_and_operations')
        new_column_names = config.get('new_column_names')
        group_by_columns = config.get("group_by_columns")

        obj = cls(columns_and_operations, new_column_names, group_by_columns)
        obj.input_block = config.get('input_block')
        return obj

    def transform(self, df):
        """
        Add a column with the maximum or minimum value to the DataFrame.

        :param df: The input DataFrame.
        :return: A new DataFrame with an additional column containing the maximum or minimum value.
        """
        window_spec = Window.partitionBy(self.group_by_columns)

        i = 0
        for input_col_name, agg_type in self.columns_and_operations.items():
            if agg_type == 'max':
                result_col = max(input_col_name).over(window_spec)
            elif agg_type == 'min':
                result_col = min(input_col_name).over(window_spec)
            
            df = df.withColumn(self.new_column_names[i], result_col)
            i+=1
        
        return df

class ImputeTransformer(Transformer):
    """
    A transformer to clean/impute missing values in specified columns.

    :param columns_to_impute: A list of column names to impute missing values.
    :param imputation_strategy: The imputation strategy. Supported values: 'mean', 'median', 'mode'.
    """

    def __init__(self, columns_to_impute, imputation_strategy):
        self.columns_to_impute = columns_to_impute
        self.imputation_strategy = imputation_strategy

    @classmethod
    def from_config(cls, config):
        columns_to_impute = config.get("columns_to_impute")
        imputation_strategy = config.get('imputation_strategy')

        obj = cls(columns_to_impute, imputation_strategy)
        obj.input_block = config.get('input_block')
        return obj

    def transform(self, df):
        """
        Add column with col_value to dataframe
        :return; Dataframe w/ new column containing col_value
        """

        imputer = Imputer(inputCols=self.columns_to_impute,
                        outputCols=["{}_imputed".format(c) for c in self.columns_to_impute]
                        ).setStrategy(self.imputation_strategy)

        df_imputed = imputer.fit(df).transform(df)
        df_imputed = df_imputed.drop(*self.columns_to_impute)
        
        # Rename back to original columns
        for c in self.columns_to_impute:
            df_imputed = df_imputed.withColumnRenamed("{}_imputed".format(c), c)

        return df_imputed

class JoinWithConditionTransformer(Transformer):
    """
    A transformer to join two DataFrames together based on a specified join condition.

    :param join_type: The type of join to perform. Currently only 'inner' join is supported.
    :param condition: The condition for the join.
    :param join_on: A dictionary where the keys are the table names (or DataFrame aliases)
                    and the values are the column names to join on for each table.
                    Example: {'table1': 'column1', 'table2': 'column2'}
    """

    def __init__(self, join_type, condition, join_on):
        self.join_type = join_type
        self.condition = condition
        self.join_on = join_on

    @classmethod
    def from_config(cls, config):
        join_type = config.get("join_type")
        condition = config.get('condition')
        join_on = config.get("join_on")

        obj = cls(join_type, condition, join_on)
        obj.input_block = config.get('input_block')
        return obj

    def transform(self, df1, df2):
        """
        Join two DataFrames together based on the specified join condition.

        :param df1: The first DataFrame to be joined.
        :param df2: The second DataFrame to be joined.
        :return: A new DataFrame resulting from the join operation.
        :raises: SerraRunException if the join condition columns do not match between the two DataFrames.
        """
        
        df1 = df1.alias('a').join(df2.alias('b'), F.expr(self.condition), self.join_type)

        return df1

class MapTransformer(Transformer):
    """
    A transformer to map values in a DataFrame column to new values based on a given mapping dictionary.

    :param config: A dictionary containing the configuration for the transformer.
                   It should have the following keys:
                   - 'output_column': The name of the new column to be added after mapping.
                   - 'mapping_dictionary': A dictionary containing the mapping of old values to new values.
                                 If 'mapping_dictionary' is not provided, 'mapping_dict_path' should be specified.
                   - 'mapping_dict_path': The path to a JSON file containing the mapping dictionary.
                   - 'input_column': The name of the DataFrame column to be used as the key for mapping.
    """
    def __init__(self, output_column, input_column, mapping_dictionary=None, mapping_dict_path=None):
        self.output_column = output_column
        self.mapping_dictionary = mapping_dictionary
        self.mapping_dict_path = mapping_dict_path
        self.input_column = input_column

    @classmethod
    def from_config(cls, config): 
        print(config)
        output_col = config.get("output_column")
        map_dict = config.get("mapping_dictionary")
        map_dict_path = config.get("mapping_dict_path")
        input_col = config.get('input_column')
        obj = cls(output_col, map_dict, map_dict_path, input_col)
        obj.input_block = config.get('input_block')
        return obj

    def transform(self, df):
        """
        Map values in the DataFrame column to new values based on the specified mapping.

        :param df: The input DataFrame to be transformed.
        :return: A new DataFrame with the new column containing the mapped values.
        :raises: SerraRunException if any required config parameter is missing or if column specified
                 as 'input_column' does not exist in the DataFrame.
        """
        if not self.output_column or not self.input_column:
            raise SerraRunException("Both 'output_column' and 'input_column' must be provided in the config.")
        
        if not self.mapping_dictionary and not self.mapping_dict_path:
            raise SerraRunException("Either 'mapping_dictionary' or 'mapping_dict_path' must be provided in the config.")

        if self.input_column not in df.columns:
            raise SerraRunException(f"Column '{self.input_column}' specified as input_column does not exist in the DataFrame.")

        if self.mapping_dictionary is None:
            with open(self.mapping_dict_path) as f:
                self.mapping_dictionary = json.load(f)

        try:
            for key, value in self.mapping_dictionary.items():
                df = df.withColumn(f'{self.output_column}_{key}', F.when(F.col(self.input_column) == key, value))

            # Select the first non-null value from the generated columns
            # create list, then unpack *
            df = df.withColumn(self.output_column, F.coalesce(*[F.col(f'{self.output_column}_{key}') for key in self.mapping_dictionary]))
            df = df.drop(*[f'{self.output_column}_{key}' for key in self.mapping_dictionary])
        except Exception as e:
            raise SerraRunException(f"Error transforming DataFrame: {str(e)}")

        return df

class MultiJoinTransformer(Transformer):
    """
    A transformer to join multiple DataFrames together based on specified join conditions.

    :param join_type: The type of join to perform. Currently only 'inner' join is supported.
    :param join_on: A dictionary where the keys are the table names (or DataFrame aliases)
                    and the values are the column names to join on for each table.
                    Example: {'table1': 'column1', 'table2': 'column2'}
    """

    def __init__(self, join_type, join_on):
        self.join_type = join_type
        self.join_on = join_on

    @classmethod
    def from_config(cls, config):
        join_type = config.get("join_type")
        join_on = config.get("join_on")

        obj = cls(join_type, join_on)
        obj.input_block = config.get('input_block')
        return obj

    def transform(self, *dfs):
        """
        Join multiple DataFrames together based on the specified join conditions.

        :param dfs: A variable number of DataFrames to be joined.
        :return: A new DataFrame resulting from the join operations.
        :raises: SerraRunException if the join condition columns do not match between the DataFrames.
        """  

        join_keys = []
        for table in self.join_on:
            join_keys.append(self.join_on.get(table))
            

        matching_col = join_keys

        joined_df = dfs[0]
        for i in range(1,len(dfs)):
            if matching_col[i-1] != matching_col[i]:
                joined_df = joined_df.join(dfs[i], joined_df[matching_col[i-1]] == dfs[i][matching_col[i]], self.join_type[i-1])
            else:
                joined_df = joined_df.join(dfs[i], matching_col[i], self.join_type[i-1])

            non_null_counts = [joined_df.where(col(c).isNotNull()).count() for c in dfs[i].columns]
            if joined_df.isEmpty() or all(count == 0 for count in non_null_counts):
                raise SerraRunException(f"""Joiner - Join Key Error: {matching_col[0]}, {matching_col[1]} columns do not match.""")

        return joined_df

class MultipleCaseWhenTransformer(Transformer):
    """
    A transformer that adds a new column to the DataFrame based on conditional rules (similar to SQL's CASE WHEN).

    :param columns_and_conditions: A list of tuples representing the conditions and their corresponding results.
                                   Each tuple should be in the format (condition_value, result_value).
                                   The condition_value can be a specific value or a pattern for LIKE comparisons.
                                   The result_value will be assigned to the output_col if the condition is met.
    :param input_column: The name of the column containing the values to be evaluated.
    :param output_col: The name of the new column to be added with the results of the conditions.
    :param type: The type of comparison to be used. It can be either '==' for equality comparison or 'like' for pattern matching using the LIKE operator.
    """

    def __init__(self, columns_and_conditions, input_column, output_col, type):
        self.columns_and_conditions = columns_and_conditions
        self.input_column = input_column
        self.output_col = output_col
        self.type = type

    @classmethod
    def from_config(cls, config):
        columns_and_conditions = config.get('columns_and_conditions')
        input_column = config.get("input_column")
        output_col = config.get('output_col')
        type = config.get('type')

        obj = cls(columns_and_conditions, input_column, output_col, type)
        obj.input_block = config.get('input_block')
        return obj

    def transform(self, df):
        """
        Add a new column with the results of the conditions to the DataFrame.

        :param df: The input DataFrame.
        :return: A new DataFrame with an additional column containing the results of the conditions.
        """
        # Mappings to indicate which comparison function to use
        type_dict = {
            '<=': lambda col, val: col <= val,
            '>=': lambda col, val: col >= val,
            '==': lambda col, val: col == val,
            'like': lambda col, val: col.like(val)
        }

        # Get the particular condition function based on the comparison type
        comparison_func = type_dict.get(self.type)
        output_cols = list(self.columns_and_conditions.keys())
        conditions = list(self.columns_and_conditions.values())

        if comparison_func is None:
            raise ValueError(f"Unsupported comparison type: {self.type}")
        
        for i in range(len(output_cols)):
            if len(conditions[i]) == 2:
                case_expr = when(comparison_func(df[self.input_column[0]], conditions[i][0]), self.parse_result_value(conditions[i][1]))
            else: 
                case_expr = when(comparison_func(df[self.input_column[0]], conditions[i][0]) | comparison_func(df[self.input_column[0]], conditions[i][2]), self.parse_result_value(conditions[i][1]))
            case_expr = case_expr.otherwise(None)
            df = df.withColumn(output_cols[i], case_expr)

        return df
    
    def parse_result_value(self, result_value):
        if 'col:' in result_value:
            return col(result_value[4:])
        else:
            return(result_value)

class OrderByTransformer(Transformer):
    """
    A transformer to sort the DataFrame based on specified columns in ascending or descending order.

    :param columns: A list of column names to sort the DataFrame by.
    :param ascending: Optional. If True (default), sort in ascending order. If False, sort in descending order.
    """

    def __init__(self, columns, ascending=True):
        self.columns = columns
        self.ascending = ascending

    @classmethod
    def from_config(cls, config):
        columns = config.get("columns")
        ascending = config.get("ascending", True)

        obj = cls(columns, ascending)
        obj.input_block = config.get('input_block')
        return obj

    def transform(self, df):
        """
        Transform the DataFrame by sorting it based on specified columns.

        :param df: The input DataFrame to be transformed.
        :return: A new DataFrame with the rows sorted based on the specified columns.
        """

        return df.orderBy(*self.columns, ascending = self.ascending)

class PivotTransformer(Transformer):
    """
    A transformer to pivot a DataFrame based on specified row and column levels, and perform aggregation.

    :param row_level_column: The column used for row levels during pivoting.
    :param column_level_column: The column used for column levels during pivoting.
    :param value_column: The column to be summarized (values) during pivoting.
    :param aggregate_type: The type of aggregation to perform after pivoting.
                           Should be one of 'avg' (average) or 'sum' (sum).
    """

    def __init__(self, row_level_column, column_level_column, value_column, aggregate_type):
        self.row_level_column = row_level_column
        self.column_level_column = column_level_column
        self.value_column = value_column
        self.aggregate_type = aggregate_type

    @classmethod
    def from_config(cls, config):
        row_level_column = config.get("row_level_column")
        column_level_column = config.get("column_level_column")
        value_column = config.get("value_column")
        aggregate_type = config.get("aggregate_type")

        obj = cls(row_level_column, column_level_column, value_column, aggregate_type)
        obj.input_block = config.get('input_block')
        return obj

    def transform(self, df):
        """
        Pivot the DataFrame based on the specified row and column levels, and perform aggregation.

        :param df: The input DataFrame to be transformed.
        :return: A new DataFrame resulting from the pivot operation with the specified aggregation.
        :raises: SerraRunException if the specified aggregation type is invalid.
        """
        
        df = df.withColumn(self.value_column, F.col(self.value_column).cast("double"))
        df = df.groupBy(self.row_level_column).pivot(self.column_level_column)

        # Perform aggregation
        if self.aggregate_type == "avg":
            df = df.avg(self.value_column)
        elif self.aggregate_type == "sum":
            df = df.sum(self.value_column)
        else:
            raise SerraRunException("Invalid Pivot Aggregation type")

        return df

class RenameColumnTransformer(Transformer):
    """
    A transformer to rename a column in a DataFrame.

    :param old_name: The name of the column to be renamed.
    :param new_name: The new name to be assigned to the column.
    """

    def __init__(self, old_name, new_name):
        self.old_name = old_name
        self.new_name = new_name

    @classmethod
    def from_config(cls, config):
        old_name = config.get("old_name")
        new_name = config.get("new_name")

        obj = cls(old_name, new_name)
        obj.input_block = config.get('input_block')
        return obj

    def transform(self, df):
        """
        Rename a column in the DataFrame.

        :param df: The input DataFrame to be transformed.
        :return: A new DataFrame with the specified column renamed.
        """

        df = df.withColumn(
            self.new_name, F.col(self.old_name)
        )
        return df.drop(F.col(self.old_name))

class SelectTransformer(Transformer):
    """
    A transformer to perform a SELECT operation on a DataFrame.

    :param columns: A list of column names to select from the DataFrame.
    :param distinct_column: Optional. The column for which distinct values should be retained.
    :param filter_expression: Optional. A filter expression to apply to the DataFrame.
    """

    def __init__(self, columns=None, distinct_column=None, filter_expression=None):
        self.columns = columns or []
        self.distinct_column = distinct_column
        self.filter_expression = filter_expression

    @classmethod
    def from_config(cls, config):
        columns = config.get("columns", [])
        distinct_column = config.get('distinct_column')
        filter_expression = config.get('filter_expression')

        obj = cls(columns, distinct_column, filter_expression)
        obj.input_block = config.get('input_block')
        return obj

    def transform(self, df: DataFrame) -> DataFrame:
        """
        Perform the SELECT operation on the DataFrame.

        :param df: The input DataFrame to be transformed.
        :return: A new DataFrame containing only the selected columns.
        :raises: SerraRunException if no columns are specified in the configuration
                 or if none of the specified columns exist in the DataFrame.
        """
        if not self.columns:
            raise SerraRunException("No columns specified in the configuration.")

        selected_columns = [F.col(col) for col in self.columns if col in df.columns]

        if not selected_columns:
            raise SerraRunException("None of the specified columns exist in the DataFrame.")

        df = df.select(*selected_columns)
        
        if self.distinct_column is not None:
            df = df.dropDuplicates(self.distinct_column)

        if self.filter_expression is not None:
            df = df.filter(F.expr(self.filter_expression))
        return df

class SQLTransformer(Transformer):
    """
    A transformer to perform a SQL SELECT operation on a DataFrame.

    :param sql_expression: A SQL expression string representing the SELECT operation.
    """

    def __init__(self, sql_expression):
        self.sql_expression = sql_expression

    @classmethod
    def from_config(cls, config):
        sql_expression = config.get('sql_expression')

        obj = cls(sql_expression)
        obj.input_block = config.get('input_block')
        return obj

    def transform(self, df: DataFrame) -> DataFrame:
        """
        Perform the SELECT operation on the DataFrame.

        :param df: The input DataFrame to be transformed.
        :return: A new DataFrame containing only the selected columns.
        :raises: SerraRunException if no columns are specified in the configuration
                 or if none of the specified columns exist in the DataFrame.
        """
        df = df.filter(F.expr(self.sql_expression))
        return df

class WindowTransformer(Transformer):
    """
    A transformer to set a window for all other steps after it.

    :param partition_by: A list of column names for partitioning the window.
    :param order_by: A list of column names for ordering within the window.
    :param window_name: The name of the window column to be added.
    """

    def __init__(self, partition_by=None, order_by=None, window_name=None):
        self.partition_by = partition_by or []
        self.order_by = order_by or []
        self.window_name = window_name

    @classmethod
    def from_config(cls, config):
        partition_by = config.get("partition_by", [])
        order_by = config.get("order_by", [])
        window_name = config.get("window_name")

        obj = cls(partition_by, order_by, window_name)
        obj.input_block = config.get('input_block')
        return obj

    def transform(self, df: DataFrame) -> DataFrame:
        window_spec = Window().partitionBy(*self.partition_by).orderBy(*self.partition_by)
        df_with_window = df.withColumn(f"window_{self.partition_by[0]}", F.row_number().over(window_spec))
        return df_with_window

class LocalWriter(Writer):
    """
    A writer to write data from a Spark DataFrame to a local file.

    :param file_path: The path of the local file to write to.
    """

    def __init__(self, file_path):
        self.file_path = file_path

    @classmethod
    def from_config(cls, config):
        file_path = config.get("file_path")
        obj = cls(file_path)
        obj.input_block = config.get('input_block')
        return obj
    
    def write(self, df):
        """
        Write data from a Spark DataFrame to a local file.

        :param df: The Spark DataFrame to be written to the local file.
        """
        # Convert PySpark DataFrame to Pandas DataFrame
        pandas_df = df.toPandas()

        # Write the Pandas DataFrame to a local file
        pandas_df.to_csv(self.file_path, index=False)

class DatabricksWriter(Writer):
    """
    A writer to write data from a Spark DataFrame to a Databricks Delta table.

    :param config: A dictionary containing the configuration for the writer.
                   It should have the following keys:
                   - 'database': The name of the Databricks database to write to.
                   - 'table': The name of the table in the Databricks database.
                   - 'format': The file format to use for the Delta table.
                   - 'mode': The write mode to use, such as 'overwrite', 'append', etc.
    """

    def __init__(self, database, table, format, mode):
        self.database = database
        self.table = table
        self.format = format
        self.mode = mode

    @classmethod
    def from_config(cls, config):
        database = config.get('database')
        table = config.get('table')
        format = config.get('format')
        mode = config.get('mode')

        obj = cls(database, table, format, mode)
        obj.input_block = config.get('input_block')
        return obj
        
    def write(self, df: DataFrame):
        """
        Write data from a Spark DataFrame to a Databricks Delta table.

        :param df: The Spark DataFrame to be written to the Delta table.
        """
        # Currently forces overwrite if csv already exists
        df.write.format(self.format).mode(self.mode).saveAsTable(f'{self.database}.{self.table}')
        return None




from serra.python.base import PythonTransformer
import pandas as pd

class RenameColumnsTransformer(PythonTransformer):
    """
    Rename specified columns in the dataset to a new column name.

    """
    def __init__(self, columns_to_change, new_column_name):
        self.columns_to_change = columns_to_change
        self.new_column_name = new_column_name
        
    def transform(self, df):
        for column in self.columns_to_change:
            df.rename(columns={column: self.new_column_name}, inplace=True)
        return df



from serra.python.base import PythonTransformer
import logging

# Configure logging at the beginning of your script or module
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')

from fuzzywuzzy import fuzz
import pandas as pd

class MatchTransformer(PythonTransformer):
    """
    Drop rows that have nulls in specified column(s).

    :param columns: The columns to look for
    """

    def __init__(self, left_df, right_df, columns_to_match, partition, include_fuzzy = True, threshold = 90):
        self.left_df = left_df
        self.right_df = right_df
        self.columns_to_match = columns_to_match # Dict: key = left column, value = right
        self.partition = partition
        self.include_fuzzy = include_fuzzy
        self.threshold = threshold
        
    def transform(self):
        print("Left DF: ", self.left_df, self.left_df[self.partition[0]])
        print("Right DF: ", self.right_df)

        matched_data = pd.DataFrame()

        for partition in self.left_df[self.partition[0]].dropna().unique():
            left_df_subset = self.left_df[self.left_df[self.partition[0]] == partition].copy()
            right_df_subset = self.right_df[self.right_df[self.partition[0]] == partition].copy()
            print("Left DF SUBSET: ", left_df_subset)
            print("Right DF SIBSET: ", right_df_subset)
            #_______Exact Matches________
            for left_column, right_column in self.columns_to_match.items():
                exact_match_name = pd.merge(left_df_subset, right_df_subset,
                                            left_on=[left_column],
                                            right_on=[right_column],
                                            how='inner',
                                            suffixes=('_left', '_right'))

                exact_match_name[left_column + '_match'] = 1
                
                matched_data = pd.concat([matched_data, exact_match_name], ignore_index=True)
                matched_data['fuzzy_score'] = None
                print(f"\n\n\n\n————————Exact Matched Data for {left_column}——————————", matched_data['fuzzy_score'], "\n\n\n\n")
            
                # Remove matched rows
                left_df_subset = left_df_subset[~left_df_subset[left_column].isin(exact_match_name[left_column])]
                right_df_subset = right_df_subset[~right_df_subset[right_column].isin(exact_match_name[right_column])]

            #______Fuzzy Matches_______
            if self.include_fuzzy:

                match_series_list = []

                    # Iterate through each left-right column match pair
                        #  Iterate through each row in left_df
                            # For each single row in left_df, iterate through each right_df row
                                # Calculate fuzzy score for each row in the right_df to the single row in left_df
                for left_column, right_column in self.columns_to_match.items():

                        
                    for index, row in left_df_subset.iterrows():

                        fuzzy_scores = []

                            
                        for index_right, row_right in right_df_subset.iterrows():

                            # Compare the single left row value to all the right_df row values and get scores
                            # Store index of right_df value and score (all in comparison to single left row value)
                            score = fuzz.token_set_ratio(row[left_column], row_right[right_column])
                            fuzzy_scores.append((index_right, score))

                        # Proceed only if there are scores to evaluate
                        if fuzzy_scores:
                            # Find the max score and corresponding index
                            max_score_index, max_score = max(fuzzy_scores, key=lambda x: x[1])

                            # If the max score is above a threshold, consider it a match
                            
                            # Row is a single column where each index is the column name and column value is the value
                            if max_score > self.threshold:
                                match_dict = {
                                    **row.to_dict(), # Dict where key = column, pair = value
                                    **right_df_subset.loc[max_score_index].to_dict(),
                                    f"{left_column}_match": -1,  # Marking the match with a special flag
                                    'fuzzy_score': max_score
                                }

                                match_series = pd.Series(match_dict, name=index) # Single column series where values are just pair from dict, where series index labels are keys from dict(columns) 
                                print(f"\n\n\n\n————————Row for {left_column}——————————", row, "\n\n\n\n")
                                print(f"\n\n\n\n————————Matched Series Dict for {left_column}——————————", match_series, "\n\n\n\n")
                                match_series_list.append(match_series.to_frame().T)  # Convert Series to DataFrame where single column w/ indexes of column names, values of pairs, signle column becomes single row where headers are column names
                                print(f"\n\n\n\n————————Transpose Match for {left_column}——————————", match_series.to_frame().T, "\n\n\n\n")
                                print(f"\n\n\n\n————————Fuzzy Matched Series List for {left_column}——————————", match_series_list, "\n\n\n\n")

                                # Remove matched row from left_df_subset and right_df_subset
                                left_df_subset = left_df_subset.drop(index)
                                right_df_subset = right_df_subset.drop(max_score_index)

                # After iterating through all columns and rows, check if there are any matches to append
                if match_series_list:  # Check if the list is not empty
                    matched_data = pd.concat([matched_data] + match_series_list, ignore_index=True)
                
                print(f"\n\n\n\n————————Joined Matched Data for {left_column}——————————", matched_data['fuzzy_score'], "\n\n\n\n")

        print("\n\n\n\n MATCHED DATA ALL: ", matched_data)
        return matched_data




# GOAL: Take config and output python script
import re

from serra.config_parser import ConfigParser
from serra.runners.graph_runner import get_order_of_execution
from serra.utils import dict_to_string


def camel_to_snake(name):
    # Use regular expressions to find capital letters and insert an underscore before them
    snake_name = re.sub(r'([A-Z])', r'_\1', name)
    # Convert to lowercase and remove leading underscore if present
    return snake_name.lstrip('_').lower()

def parse_config(cf):
    # Returns list of each step name, class, and argument
    ordered_block_names = get_order_of_execution(cf)
    ordered_block_classes = []

    for block_name in ordered_block_names:
        block_class = cf.get_class_name_for_step(block_name)
        block_arguments = dict_to_string(cf.get_config_for_block(block_name))

        ordered_block_classes.append((block_name,block_class,block_arguments))

    return ordered_block_classes

def generate_py_script(ordered_block_classes, cf):
    with open('serra_files.py', 'r') as file:
        script = [file.read()]

    script.append("from pyspark.sql import SparkSession")
    script.append("""spark = SparkSession.builder.appName("PyRun").master("local").getOrCreate()""")

    reader_count = 0 
    join_dict = {}
    for block_name, class_name, args in ordered_block_classes:

        if "Reader" in class_name:
            line = f"df{reader_count + 1} = {class_name}({args}).read_with_spark(spark)"

            # join_dict{step_name: dfx}
            join_dict[block_name] = f"df{reader_count + 1}"
            reader_count += 1

        if "Writer" in class_name:
            line = f"{class_name}({args}).write(df1)"
        if "Transformer" in class_name:
            line = f"df1 = {class_name}({args})."
            if "Join" in class_name:
                # Go into join_on = {read_sales: id, read_ratings: id}, match the keys to the join_dict so you get the appropriate df
                join_args = cf.get_config_for_block(block_name)['join_on']
                matching_keys = [key for key in join_args if key in join_dict]
                join_dfs = [str(join_dict[key]) for key in matching_keys]
                join_dfs_string = ','.join(join_dfs)

                line += f"transform({join_dfs_string})"
            else:
                line += "transform(df1)"


        script.append(line)

    script.append("df1.show()")
    python_script = "\n".join(script)
    return python_script

def save_file(file_name, python_script):
    with open(file_name, 'w') as script_file:
        script_file.write(python_script)

    print(f"The Python script has been saved to {file_name}")

def create_and_save_py_script(config_path):
    cf = ConfigParser.from_local_config(config_path)
    ords = parse_config(cf)
    py_script = generate_py_script(ords,cf)
    save_file("py_script.py", py_script)

if __name__ == "__main__":
    create_and_save_py_script("./jobs/DBDemo.yml")



from .arrange_transformer import ArrangeTransformer
from .coalese_prep_transformer import CoalesceTransformer
from .count_rows_transformer import CountRowsTransformer
from .data_cleansing_prep_transformer import DataCleansingTransformer
from .drop_duplicates_prep_transformer import DropDuplicatesTransformer
from .drop_null_rows_prep_transformer import DropNullRowsTransformer
from .filter_rows_prep_transformer import FilterRowsTransformer
from .formula_transformer import FormulaTransformer
from .fuzzy_match_transformer import FuzzyMatchTransformer
from .join_transformer import JoinTransformer
from .match_transformer import MatchTransformer
from .running_totals_transformer import RunningTotalsTransformer
from .sample_prep_transformer import SampleTransformer
from .select_columns_prep_transformer import SelectColumnsTransformer
from .sort_prep_transformer import SortTransformer
from .summarize_transformer import SummarizeTransformer
from .tokenize_transformer import TokenizeTransformer
from .transpose_transformer import TransposeTransformer
from .unique_rows_prep_transformer import UniqueTransformer



from pyspark.ml.feature import Imputer
from pyspark.sql import functions as F
from serra.transformers.transformer import Transformer

class ImputeTransformer(Transformer):
    """
    A transformer to clean/impute missing values in specified columns.

    :param columns_to_impute: A list of column names to impute missing values.
    :param imputation_strategy: The imputation strategy. Supported values: 'mean', 'median', 'mode'.
    """

    def __init__(self, columns_to_impute, imputation_strategy):
        self.columns_to_impute = columns_to_impute
        self.imputation_strategy = imputation_strategy

    def transform(self, df):
        """
        Add column with col_value to dataframe
        :return; Dataframe w/ new column containing col_value
        """

        imputer = Imputer(inputCols=self.columns_to_impute,
                        outputCols=["{}_imputed".format(c) for c in self.columns_to_impute]
                        ).setStrategy(self.imputation_strategy)

        df_imputed = imputer.fit(df).transform(df)
        df_imputed = df_imputed.drop(*self.columns_to_impute)
        
        # Rename back to original columns
        for c in self.columns_to_impute:
            df_imputed = df_imputed.withColumnRenamed("{}_imputed".format(c), c)

        return df_imputed



from abc import abstractmethod
from serra.base import Step

class Writer(Step):
    """
    Writer base class for loading data
    Enforce write method
    """
    @abstractmethod
    def write(self, df):
        pass

    @classmethod
    def from_config(cls, config):
        c = dict(config)
        input_block = c.pop("input_block")

        obj = cls(**c)
        obj.input_block = input_block
        return obj

    @property
    def dependencies(self):
        return [self.input_block]


from serra.readers import Reader


class BigQueryReader(Reader):
    """
    A reader to read data from BigQuery into a Spark DataFrame.

    :param config: A dictionary containing the configuration for the reader.
                   It should have the following keys:
                   - 'project'
                   - 'dataset'
                   - 'table'
    """

    def __init__(self, project, dataset, table, mode):
        self.project = project
        self.dataset = dataset
        self.table = table
        self.mode = mode

    def read(self):
        """
        Read data from Snowflake and return a Spark DataFrame.

        :return: A Spark DataFrame containing the data read from the specified Snowflake table.
        """
        df = self.spark.read.format("bigquery")\
            .option('parentProject', self.project)\
            .load(f"{self.dataset}.{self.table}")
        return df



import pandas as pd

from serra.python.base import PythonTransformer

class RunningTotalsTransformer(PythonTransformer):
    """
    Create new columns of cumulative sums for given columns. Can group by and sort. 

    :param columns: Columns to calculate sums.
    :param group_by: Column(s) to group by.
    :param sort_by: Column(s) to sort by.
    """

    def __init__(self, columns, group_by = None, sort_by = None):
        self.columns = columns
        self.group_by = group_by
        self.sort_by = sort_by

    def transform(self, df):
        """
        First, create cumulative column names with list comphrension.
        Second, use these new column names to assign the cumulative sums to (group by first if needed).
        Third, sort.
        """
        cum_columns = ['cum_' + col for col in self.columns]

        if self.group_by != None:
            df[cum_columns] = df.groupby(self.group_by)[self.columns].cumsum()
        else:
            df[cum_columns] = df[self.columns].cumsum()

        if self.sort_by != None:
            df = df.sort_values(by = self.sort_by)

        return df


from serra.python.base import PythonTransformer
import string
import re

class DropDuplicatesTransformer(PythonTransformer):
    """
    Clean whitespaces, cases, punctuation.

    """

    def __init__(self, columns):
        self.columns = columns
        
    def transform(self, df):
        """
        Get mapping dictionaries. Keys are whatever use case you want, Values are dictionaries with keys as column name and values as the functions that do that specific use case transform.
        Loop through key-value pairs for the specified dictionaries, assign df[key] = value.

        """

        return df.drop_duplicates(subset=self.columns)





from pyspark.sql import functions as F

from serra.exceptions import SerraRunException
from serra.transformers.transformer import Transformer

class JoinTransformer(Transformer):
    """
    A transformer to join two DataFrames together based on a specified join condition.

    :param config: A dictionary containing the configuration for the transformer.
                   It should have the following keys:
                   - 'join_type': The type of join to perform. Currently only 'inner' join is supported.
                   - 'join_on': A dictionary where the keys are the table names (or DataFrame aliases)
                                and the values are the column names to join on for each table.
                   Example: {'table1': 'column1', 'table2': 'column2'}
    """


    def __init__(self, join_type, join_on):
        self.join_type = join_type
        self.join_on = join_on
    
    @property
    def dependencies(self):
        return self.input_block

    def transform(self, df1, df2):
        """
        Join two DataFrames together based on the specified join condition.

        :param df1: The first DataFrame to be joined.
        :param df2: The second DataFrame to be joined.
        :return: A new DataFrame resulting from the join operation.
        :raises: SerraRunException if the join condition columns do not match between the two DataFrames.
        """
        # assert self.join_type in "inner"

        join_keys = []

        
        for table in self.join_on:
            join_keys.append(self.join_on.get(table))

        assert len(join_keys) == 2

        matching_col = join_keys

        df1 = df1.join(df2, df1[matching_col[0]] == df2[matching_col[1]], self.join_type).drop(df2[matching_col[1]])
        if df1.isEmpty():
            raise SerraRunException(f"""Joiner - Join Key Error: {matching_col[0]}, {matching_col[1]} columns do not match.""")
        return df1
    
    


import boto3
from serra.python.base import PythonReader
import pandas as pd
from io import StringIO

class S3Reader(PythonReader):
    def __init__(self, aws_access_key_id: str, aws_secret_access_key: str, bucket_name: str, path: str, region_name: str):
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.bucket_name = bucket_name
        self.path = path
        self.region_name = region_name
    
    def read(self):
        s3 = boto3.client('s3', aws_access_key_id=self.aws_access_key_id, aws_secret_access_key=self.aws_secret_access_key, region_name=self.region_name)
        obj = s3.get_object(Bucket=self.bucket_name, Key=self.path)
        stream = obj['Body']
        data = stream.read().decode('utf-8')
        df = pd.read_csv(StringIO(data))
        return df


from pyspark.sql import DataFrame
from pyspark.sql import functions as F

from serra.transformers.transformer import Transformer

class SQLTransformer(Transformer):
    """
    A transformer to perform a SQL SELECT operation on a DataFrame.

    :param sql_expression: A SQL expression string representing the SELECT operation.
    """

    def __init__(self, sql_expression):
        self.sql_expression = sql_expression

    def transform(self, df: DataFrame) -> DataFrame:
        """
        Perform the SELECT operation on the DataFrame.

        :param df: The input DataFrame to be transformed.
        :return: A new DataFrame containing only the selected columns.
        :raises: SerraRunException if no columns are specified in the configuration
                 or if none of the specified columns exist in the DataFrame.
        """
        df = df.filter(F.expr(self.sql_expression))
        return df





from pyspark.sql import DataFrame

from serra.writers import Writer

class DatabricksWriter(Writer):
    """
    A writer to write data from a Spark DataFrame to a Databricks Delta table.

    :param config: A dictionary containing the configuration for the writer.
                   It should have the following keys:
                   - 'database': The name of the Databricks database to write to.
                   - 'table': The name of the table in the Databricks database.
                   - 'format': The file format to use for the Delta table.
                   - 'mode': The write mode to use, such as 'overwrite', 'append', etc.
    """

    def __init__(self, database, table, format, mode):
        self.database = database
        self.table = table
        self.format = format
        self.mode = mode
        
    def write(self, df: DataFrame):
        """
        Write data from a Spark DataFrame to a Databricks Delta table.

        :param df: The Spark DataFrame to be written to the Delta table.
        """
        # Currently forces overwrite if csv already exists
        df.write.format(self.format).mode(self.mode).saveAsTable(f'{self.database}.{self.table}')
        return None




from serra.readers import Reader

class SnowflakeReader(Reader):
    """
    A reader to read data from Snowflake into a Spark DataFrame.

    :param config: A dictionary containing the configuration for the reader.
                   It should have the following keys:
                   - 'warehouse': The Snowflake warehouse to use for the connection.
                   - 'database': The Snowflake database to use for the connection.
                   - 'schema': The Snowflake schema to use for the connection.
                   - 'table': The name of the table to be read from Snowflake.
    """

    def __init__(self, warehouse, database, schema, table, user, password, host):
        self.warehouse = warehouse
        self.database = database
        self.schema = schema
        self.table = table
        self.user = user
        self.password = password
        self.host = host
    
    def read(self):
        """
        Read data from Snowflake and return a Spark DataFrame.

        :return: A Spark DataFrame containing the data read from the specified Snowflake table.
        """
        snowflake_table = (self.spark.read
        .format("snowflake")
        .option("host", self.host)
        .option("user", self.user)
        .option("password", self.password)
        .option("sfWarehouse", self.warehouse)
        .option("database", self.database)
        .option("schema", self.schema) # Optional - will use default schema "public" if not specified.
        .option("autopushdown", "off")
        .option("dbtable", self.table)
        .load()
        )
        return snowflake_table


import os
import json

import boto3

from serra.utils import get_local_serra_profile

def get_s3_client():
    user_profile = get_local_serra_profile()
    s3 = boto3.resource('s3',
            aws_access_key_id=user_profile.aws_access_key_id,
            aws_secret_access_key=user_profile.aws_secret_access_key)
    return s3

def upload_file_to_config_bucket(file_path):
    bucket = get_local_serra_profile().aws_config_bucket
    upload_file_to_bucket(file_path, bucket)

def retrieve_file_from_config_bucket(filename):
    bucket = get_local_serra_profile().aws_config_bucket
    data = retrieve_file_as_bytes_from_bucket(filename, bucket)
    return data

# General AWS helper fxns
def upload_file_to_bucket(file_path, bucket):
    name = file_path.split("/")[-1]
    with open(file_path, 'rb') as data:
        get_s3_client().Bucket(bucket).put_object(Key=name, Body=data)

def retrieve_file_as_bytes_from_bucket(filename, bucket):
    obj = get_s3_client().Object(bucket_name=bucket, key=filename)
    response = obj.get()
    data = response['Body'].read()
    return data

def read_json_s3(file, bucket):
    s3 = get_s3_client()
    content_object = s3.Object(bucket, f'{file}.json')
    file_content = content_object.get()['Body'].read().decode('utf-8')
    return json.loads(file_content)

def write_json_s3(obj, file, bucket):
    s3 = get_s3_client()
    json_dump_s3 = lambda obj, f: s3.Object(bucket, key=f).put(Body=json.dumps(obj))
    json_dump_s3(obj, f'{file}.json')

def copy_folder_to_s3(local_folder_path, bucket_name, s3_folder_key):
    """Copy a local folder and its contents to an S3 bucket."""
    s3_client = boto3.client("s3")

    for root, dirs, files in os.walk(local_folder_path):
        for file in files:
            local_file_path = os.path.join(root, file)
            relative_path = os.path.relpath(local_file_path, local_folder_path)
            s3_file_key = os.path.join(s3_folder_key, relative_path)
            s3_client.upload_file(local_file_path, bucket_name, s3_file_key)




from serra.transformers import Transformer
from pyspark.sql import DataFrame
from pyspark.sql.functions import lit

class RowCountTransformer(Transformer):
    """
    A transformer to add a column with the total row count of the DataFrame.
    """

    def __init__(self, output_column):
        """
        :param output_column: The name of the new column to store the row count.
        """
        self.output_column = output_column

    def transform(self, df: DataFrame) -> DataFrame:
        """
        Add a column with the total row count to the DataFrame.

        :param df: The input DataFrame to be transformed.
        :return: A new DataFrame with an additional column containing the row count.
        """
        row_count = df.count()
        return df.withColumn(self.output_column, lit(row_count))


from serra.transformers import Transformer
from pyspark.sql import DataFrame

class DropDuplicateRowsTransformer(Transformer):
    """
    A transformer to drop duplicate rows in a DataFrame.
    """

    def __init__(self):
        pass

    def transform(self, df: DataFrame) -> DataFrame:
        """
        Drop duplicate rows in the DataFrame.

        :param df: The input DataFrame to be transformed.
        :return: A new DataFrame with duplicate rows removed.
        """
        return df.dropDuplicates()


from serra.transformers.transformer import Transformer

class OrderByTransformer(Transformer):
    """
    A transformer to sort the DataFrame based on specified columns in ascending or descending order.

    :param columns: A list of column names to sort the DataFrame by.
    :param ascending: Optional. If True (default), sort in ascending order. If False, sort in descending order.
    """

    def __init__(self, columns, ascending=True):
        self.columns = columns
        self.ascending = ascending

    def transform(self, df):
        """
        Transform the DataFrame by sorting it based on specified columns.

        :param df: The input DataFrame to be transformed.
        :return: A new DataFrame with the rows sorted based on the specified columns.
        """

        return df.orderBy(*self.columns, ascending = self.ascending)



from serra.python.base import PythonTransformer
import pandas as pd
import numpy as np

class FormulaTransformer(PythonTransformer):
    """
    Allows business users to write their own math transforms
    """

    def __init__(self, new_column_name, if_condition, then_statement, else_statement):
        self.new_column_name = new_column_name
        self.if_condition = if_condition
        self.then_statement = then_statement
        self.else_statement = else_statement

    def transform(self, df):
        def is_valid_eval_output(output):
            return type(output) == pd.Series or type(output) == bool or output == None


        # Evaluate expressions and ensure they are of the correct type
        eval_if_condition = df.eval(self.if_condition)
        eval_then_condition = df.eval(self.then_statement)
        eval_else_condition = df.eval(self.else_statement)
        # assert is_valid_eval_output(eval_if_condition)
        # assert is_valid_eval_output(eval_then_condition)
        # assert is_valid_eval_output(eval_else_condition)

        out = np.where(eval_if_condition, eval_then_condition, eval_else_condition)
        df[self.new_column_name] = out
        return df







import setuptools
from setuptools import setup

setup(name='serra',
      version='0.7.194',
      description='Simplified Data Pipelines',
      url='http://github.com',
      author='Serra Technologies',
      author_email='founders@serra.io',
      long_description="Placeholder",
      packages=setuptools.find_packages(),
      package_data={"serra": ["data/workspace_example/*",
                              "data/workspace_example/*/*",
                              "data/autocomplete/*",
                              "data/autocomplete/*/*"]},
      install_requires=[
          "click",
          "pyspark",
          "pyyaml", 
          "pandas", 
          "boto3", 
          "databricks-sdk",
          "wheel",
          "loguru",
          "snowflake-connector-python",
          "google-cloud-bigquery",
          "geopy",
          "db-dtypes"
      ],
      zip_safe=False,
      entry_points={
        'console_scripts': [
            'serra=serra.cli:main',
            'serra_databricks=serra.cli:serra_databricks'
        ]
    })



from pyspark.sql import SparkSession
from serra.profile import SerraProfile


def misc_config(spark_builder):
    return spark_builder.config("spark.sql.debug.maxToStringFields", 100)

def add_s3_config(spark_builder, serra_profile: SerraProfile):
     s3_access_key = serra_profile.aws_access_key_id
     s3_secret_key = serra_profile.aws_secret_access_key

     builder = (
        spark_builder.config('spark.hadoop.fs.s3a.access.key', s3_access_key)
        .config('spark.hadoop.fs.s3a.secret.key', s3_secret_key)
        .config('spark.hadoop.fs.s3a.aws.credentials.provider', 'org.apache.hadoop.fs.s3a.SimpleAWSCredentialsProvider')
     )
     return builder

def set_jar_packages(spark_builder):
    packages = [
        # for s3
        "org.apache.hadoop:hadoop-aws:3.3.4",
        "com.amazonaws:aws-java-sdk-bundle:1.12.530",
        "org.apache.hadoop:hadoop-common:3.3.4",
        # For bigquery
        "com.google.cloud.spark:spark-bigquery-with-dependencies_2.12:0.32.2",
        # For mongo db
        "org.mongodb.spark:mongo-spark-connector_2.12:10.2.1"
        # For snowflake
        # "net.snowflake:spark-snowflake_2.13:2.13.0-spark_3.4"
        ]  
    spark_builder = spark_builder.config('spark.jars.packages', ','.join(packages))
    return spark_builder

def get_or_create_spark_session(serra_profile: SerraProfile):
    # TODO: Take a look for spark session conf: https://engineeringfordatascience.com/posts/pyspark_unit_testing_with_pytest/
    builder = SparkSession.builder
    builder = set_jar_packages(builder)
    builder = add_s3_config(builder, serra_profile)
    builder = misc_config(builder)
    return builder.getOrCreate()



from pyspark.sql import functions as F

from serra.transformers.transformer import Transformer

class FilterTransformer(Transformer):
    """
    A transformer to filter the DataFrame based on a specified condition.

    :param filter_values: A list of values to filter the DataFrame by.
    :param filter_column: The name of the column to apply the filter on.
    :param is_expression: Whether the filter is an expression (boolean condition).
    """

    def __init__(self, filter_values, filter_column, is_expression=False):
        self.filter_values = filter_values
        self.filter_column = filter_column
        self.is_expression = is_expression

    def transform(self, df):
        """
        Filter the DataFrame based on the specified condition.

        :param df: The input DataFrame to be transformed.
        :return: A new DataFrame with rows filtered based on the specified condition.
        """
        if self.is_expression:
            return df.filter(F.expr(self.filter_values))

        return df.filter(df[self.filter_column].isin(self.filter_values))


from pyspark.sql.functions import col

from serra.exceptions import SerraRunException
from serra.transformers.transformer import Transformer

class MultiJoinTransformer(Transformer):
    """
    A transformer to join multiple DataFrames together based on specified join conditions.

    :param join_type: The type of join to perform. Currently only 'inner' join is supported.
    :param join_on: A dictionary where the keys are the table names (or DataFrame aliases)
                    and the values are the column names to join on for each table.
                    Example: {'table1': 'column1', 'table2': 'column2'}
    """

    def __init__(self, join_type, join_on):
        self.join_type = join_type
        self.join_on = join_on

    def transform(self, *dfs):
        """
        Join multiple DataFrames together based on the specified join conditions.

        :param dfs: A variable number of DataFrames to be joined.
        :return: A new DataFrame resulting from the join operations.
        :raises: SerraRunException if the join condition columns do not match between the DataFrames.
        """  

        join_keys = []
        for table in self.join_on:
            join_keys.append(self.join_on.get(table))
            

        matching_col = join_keys

        joined_df = dfs[0]
        for i in range(1,len(dfs)):
            if matching_col[i-1] != matching_col[i]:
                joined_df = joined_df.join(dfs[i], joined_df[matching_col[i-1]] == dfs[i][matching_col[i]], self.join_type[i-1])
            else:
                joined_df = joined_df.join(dfs[i], matching_col[i], self.join_type[i-1])

            non_null_counts = [joined_df.where(col(c).isNotNull()).count() for c in dfs[i].columns]
            if joined_df.isEmpty() or all(count == 0 for count in non_null_counts):
                raise SerraRunException(f"""Joiner - Join Key Error: {matching_col[0]}, {matching_col[1]} columns do not match.""")

        return joined_df



from serra.python.base import PythonTransformer
class SummarizeTransformer(PythonTransformer):
    """
    This class is a transformer that takes in a dataframe and returns a dataframe with the summary statistics
    of the input dataframe. The summary statistics are defined by the user and are passed in as a list of dictionaries.
    Each dictionary contains the column name and the summary statistic to be calculated. The summary statistics are
    calculated using the pandas library.
    summary_transforms: list of dictionaries
        format: [{"column": "column_name", "function": "summary_statistic"}, ...]

    the summary_statistic can be one of: "Sum", "Mean", "Min", "Max", "Concatenate"

    summary_transforms example: 
    [
            {"column": "school", 'summary_function': "Group By"},
            {"column": "points", 'summary_function': "Sum"},
            {"column": "points", 'summary_function': "Mean"},
            {"column": "points", 'summary_function': "Min"},
            {"column": "points", 'summary_function': "Max"},
            {"column": "Name", 'summary_function': "Concatenate"}
    ]
    """

    def __init__(self, summary_transforms):
        self.summary_transforms = summary_transforms

    def transform(self, df):
        example_config = self.summary_transforms

        def concatenate(x):
            return ", ".join(x.values)

        convert_selected_function_to_function = {
            "Sum": "sum",
            "Mean": "mean",
            "Min": "min",
            "Max": "max",
            "Concatenate": concatenate
        }

        # Configure the input for the pandas group by
        group_by_columns = []
        for config in example_config:
            if config['summary_function'] == "Group By":
                group_by_columns.append(config['column'])

        # Configure the input for aggregation for pandas
        agg_input = {}
        for config in example_config:
            if config['summary_function'] == "Group By":
                continue
            
            # Get converted function
            func = convert_selected_function_to_function[config['summary_function']]
            
            if config['column'] in agg_input:
                agg_input[config['column']].append(func)
            else:
                agg_input[config['column']] = [func]

        # TODO: Verify it's being used properly
        number_functions = ["Sum", "Mean", "Min", "Max"]
        string_functions = ["Concatenate"]

        another_group = df.groupby(group_by_columns)
        final = another_group.agg(agg_input)

        # Concatenate 
        final.columns = ["_".join(a) for a in final.columns.to_flat_index()]

        # Convert pandas dataframe with multiindex to single index
        final.reset_index(inplace=True)

        return final



from pyspark.sql import functions as F

from serra.transformers.transformer import Transformer

class AggregateTransformer(Transformer):
    """
    A transformer to aggregate data based on specified columns and aggregation type.

    :param group_by_columns: A list of column names to group the DataFrame by.
    :param aggregation_type: The type of aggregation to be performed. Supported values: 'sum', 'avg', 'count'.
    """

    def __init__(self, group_by_columns, aggregation_type):
        self.group_by_columns = group_by_columns
        self.aggregation_type = aggregation_type

    def transform(self, df):
        """
        Transform the DataFrame by aggregating data based on the specified configuration.

        :param df: The input DataFrame to be transformed.
        :return: A new DataFrame with the aggregated data.
        """
        df = df.groupBy(*self.group_by_columns)

        if self.aggregation_type == 'sum':
            df = df.sum()
        if self.type == 'avg':
            df = df.mean()
        if self.type == 'count':
            df = df.count()

        return df



from .add_column_transformer import AddColumnTransformer
from .case_when_transformer import CaseWhenTransformer
from .cast_columns_transformer import CastColumnsTransformer
from .drop_columns_transformer import DropColumnsTransformer
from .get_count_transformer import GetCountTransformer
from .join_transformer import JoinTransformer
from .map_transformer import MapTransformer
from .pivot_transformer import PivotTransformer
from .rename_column_transformer import RenameColumnTransformer
from .select_transformer import SelectTransformer
from .filter_transformer import FilterTransformer
from .window_transformer import WindowTransformer
from .multi_join_transformer import MultiJoinTransformer
from .date_trunc_transformer import DateTruncTransformer
from .join_with_condition_transformer import JoinWithConditionTransformer
from .geo_distance_transformer import GeoDistanceTransformer
from .cross_join_transformer import CrossJoinTransformer
from .case_when_transformer import CaseWhenTransformer
from .multiple_case_when_transformer import MultipleCaseWhenTransformer
from .get_max_or_min_transformer import GetMaxOrMinTransformer
from .custom_transformer import CustomTransformer
from .custom_transformer import CustomTransformer
from .get_distinct_transformer import GetDistinctTransformer
from .de_duplicate_transformer import DeDuplicateTransformer
from .drop_duplicate_rows_transformer import DropDuplicateRowsTransformer
from .row_count_transformer import RowCountTransformer
from .drop_null_transformer import DropNullTransformer
from .transformer import Transformer


from loguru import logger

from serra.writers import Writer

class SnowflakeWriter(Writer):
    """
    A writer to write data from a Spark DataFrame to a Snowflake table.

    :param config: A dictionary containing the configuration for the writer.
                   It should have the following keys:
                   - 'type': The type of operation to perform. Possible values are 'create' or 'insert'.
                   - 'warehouse': The Snowflake warehouse to use for the connection.
                   - 'database': The name of the Snowflake database to write to.
                   - 'schema': The name of the Snowflake schema to write to.
                   - 'table': The name of the Snowflake table to write to.
    """

    def __init__(self, warehouse, database, schema, table, host, user, password, mode):
        self.warehouse = warehouse
        self.database = database
        self.schema = schema
        self.table = table
        self.host = host
        self.user = user
        self.password = password
        self.mode = mode
    
    def write(self, df):
        """
        Write data from a Spark DataFrame to a Snowflake table.
        :param df: The Spark DataFrame to be written to the Snowflake table.
        """
        assert self.mode in ['append', 'overwrite', 'error', 'ignore']

        (df.write
        .format("snowflake") 
        .option("host", self.host)
        .option("user", self.user)
        .option("password", self.password)
        .option("sfWarehouse", self.warehouse)
        .option("database", self.database)
        .option("schema", self.schema)
        .option("dbtable", self.table) 
        .option("autopushdown", "off")
        .mode(self.mode) 
        .save()
        )



import json

def read_json_file(file_path):
    # Opening JSON file
    data = {}
    with open(file_path) as json_file:
        data = json.load(json_file)

    return data

def write_to_file(data_dict, output_file):
    try:
        with open(output_file, 'w') as file:
            json.dump(data_dict, file, indent=4, sort_keys=True)
        print(f"Dictionary has been pretty printed to '{output_file}' successfully.")
    except Exception as e:
        print(f"Error occurred while writing to file: {e}")

SPEC_PATH = "./inputs/specs.json"
SCHEMA_TEMPLATE = "./templates/schema_template.json"
BLOCK_TEMPLATE = "./templates/transformer_template.json"
OUTPUT_DIR = "./output"

specs = read_json_file(SPEC_PATH)
schema_template = read_json_file(SCHEMA_TEMPLATE)
transformer_template = read_json_file("./templates/transformer_template.json")

# Create the main schema json
for class_name in specs:
    new_property = {
            "type": "object",
            "properties": {
                f"{class_name}" : { "$ref": f"./{class_name}.json" },
                "tests": {"type": "array"}
                },
            "additionalProperties": False
          }

    schema_template["patternProperties"]["^(?!debug$).*"]['oneOf'].append(new_property)

# Create the schemas for each class
for class_name in specs:
    class_schema = read_json_file(BLOCK_TEMPLATE)

    for property,property_type in specs[class_name].items():
        class_schema['properties'][property] = {"type": property_type}
    write_to_file(class_schema, f"{OUTPUT_DIR}/{class_name}.json")

write_to_file(schema_template,f"{OUTPUT_DIR}/serra_yaml_schema.json")



from pyspark.sql import functions as F

from serra.transformers.transformer import Transformer

class DropColumnsTransformer(Transformer):
    """
    A transformer to drop specified columns from the DataFrame.

    :param config: A dictionary containing the configuration for the transformer.
                   It should have the following key:
                   - 'columns_to_drop': A list of column names to be dropped from the DataFrame.
    """

    def __init__(self, columns_to_drop):
        self.columns_to_drop = columns_to_drop

    def transform(self, df):
        """
        Drop specified columns from the DataFrame.

        :param df: The input DataFrame to be transformed.
        :return: A new DataFrame with the specified columns dropped.
        """
        return df.select([c for c in df.columns if c not in self.columns_to_drop])
    


from pyspark.sql import functions as F
import json
from serra.transformers.transformer import Transformer
from serra.exceptions import SerraRunException

class MapTransformer(Transformer):
    """
    A transformer to map values in a DataFrame column to new values based on a given mapping dictionary.

    :param config: A dictionary containing the configuration for the transformer.
                   It should have the following keys:
                   - 'output_column': The name of the new column to be added after mapping.
                   - 'mapping_dictionary': A dictionary containing the mapping of old values to new values.
                                 If 'mapping_dictionary' is not provided, 'mapping_dict_path' should be specified.
                   - 'mapping_dict_path': The path to a JSON file containing the mapping dictionary.
                   - 'input_column': The name of the DataFrame column to be used as the key for mapping.
    """
    def __init__(self, output_column, input_column, mapping_dictionary=None, mapping_dict_path=None):
        self.output_column = output_column
        self.mapping_dictionary = mapping_dictionary
        self.mapping_dict_path = mapping_dict_path
        self.input_column = input_column

    def transform(self, df):
        """
        Map values in the DataFrame column to new values based on the specified mapping.

        :param df: The input DataFrame to be transformed.
        :return: A new DataFrame with the new column containing the mapped values.
        :raises: SerraRunException if any required config parameter is missing or if column specified
                 as 'input_column' does not exist in the DataFrame.
        """
        if not self.output_column or not self.input_column:
            raise SerraRunException("Both 'output_column' and 'input_column' must be provided in the config.")
        
        if not self.mapping_dictionary and not self.mapping_dict_path:
            raise SerraRunException("Either 'mapping_dictionary' or 'mapping_dict_path' must be provided in the config.")

        if self.input_column not in df.columns:
            raise SerraRunException(f"Column '{self.input_column}' specified as input_column does not exist in the DataFrame.")

        if self.mapping_dictionary is None:
            with open(self.mapping_dict_path) as f:
                self.mapping_dictionary = json.load(f)

        try:
            for key, value in self.mapping_dictionary.items():
                df = df.withColumn(f'{self.output_column}_{key}', F.when(F.col(self.input_column) == key, value))

            # Select the first non-null value from the generated columns
            # create list, then unpack *
            df = df.withColumn(self.output_column, F.coalesce(*[F.col(f'{self.output_column}_{key}') for key in self.mapping_dictionary]))
            df = df.drop(*[f'{self.output_column}_{key}' for key in self.mapping_dictionary])
        except Exception as e:
            raise SerraRunException(f"Error transforming DataFrame: {str(e)}")

        return df
